/* 
 * Copyright (C) 2001-2005 Jacek Sieka, arnetheduck on gmail point com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "stdinc.h"
#include "DCPlusPlus.h"

#include "FinishedManager.h"

FinishedManager::~FinishedManager() throw() {
	DownloadManager::getInstance()->removeListener(this);
	UploadManager::getInstance()->removeListener(this);
	{
		Lock l(cs);
		for_each(downloads.begin(), downloads.end(), DeleteFunction<FinishedItem*>());
		for_each(uploads.begin(), uploads.end(), DeleteFunction<FinishedItem*>());
	}
}

void FinishedManager::remove(FinishedItem *item, bool upload /* = false */) {
	{
		Lock l(cs);
		FinishedItem::List *listptr = upload ? &uploads : &downloads;
		FinishedItem::Iter it = find(listptr->begin(), listptr->end(), item);

		if(it != listptr->end())
			listptr->erase(it);
		else
			return;
	}
	if (!upload)
		fire(FinishedManagerListener::RemovedDl(), item);
	else
		fire(FinishedManagerListener::RemovedUl(), item);
	delete item;		
}

void FinishedManager::removeAll(bool upload /* = false */) {
	{
		Lock l(cs);
		FinishedItem::List *listptr = upload ? &uploads : &downloads;
		for_each(listptr->begin(), listptr->end(), DeleteFunction<FinishedItem*>());
		listptr->clear();
	}
	if (!upload)
		fire(FinishedManagerListener::RemovedAllDl());
	else
		fire(FinishedManagerListener::RemovedAllUl());
}

void FinishedManager::on(DownloadManagerListener::Complete, Download* d) throw()
{
	if((!d->isSet(Download::FLAG_USER_LIST) || BOOLSETTING(LOG_FILELIST_TRANSFERS)) && !d->isSet(Download::FLAG_TREE_DOWNLOAD)) {
		FinishedItem *item = new FinishedItem(
			d->getTarget(), d->getUserConnection()->getUser()->getNick(),
			d->getUserConnection()->getUser()->getLastHubName(),
			d->getSize(), d->getTotal(), (GET_TICK() - d->getStart()), GET_TIME(), 
			d->isSet(Download::FLAG_CRC_TTH_OK), d->isSet(Download::FLAG_CRC_TTH_BAD));
		{
			Lock l(cs);
			downloads.push_back(item);
		}
	
		fire(FinishedManagerListener::AddedDl(), item);
	}
	if(BOOLSETTING(REPORT_DOWNLOAD) && d->isSet(Download::FLAG_TREE_DOWNLOAD) == false) {
		char* buf = new char[STRING(FINISHED_DOWNLOAD).size() + MAX_PATH + 128];
		sprintf(buf, CSTRING(FINISHED_DOWNLOAD), d->getTargetFileName().c_str(), 
			d->getUserConnection()->getUser()->getNick().c_str());
		LogManager::getInstance()->message(buf, false);
		delete[] buf;
	}
}

void FinishedManager::on(UploadManagerListener::Complete, Upload* u) throw()
{
	if((!u->isSet(Upload::FLAG_USER_LIST) || BOOLSETTING(LOG_FILELIST_TRANSFERS)) && u->isSet(Upload::FLAG_TTH_LEAVES) == false &&
		(BOOLSETTING(LOG_PARTIAL_UPLOADS) || u->getUserConnection()->getEndOnEndFile() == true)) {
		FinishedItem *item = new FinishedItem(
			u->getLocalFileName(), u->getUserConnection()->getUser()->getNick(),
			u->getUserConnection()->getUser()->getLastHubName(),
			u->getSize(), u->getTotal(), (GET_TICK() - u->getStart()), GET_TIME());
		{
			Lock l(cs);
			uploads.push_back(item);
		}
			
		fire(FinishedManagerListener::AddedUl(), item);
	}
	if(BOOLSETTING(REPORT_UPLOAD) && u->isSet(Upload::FLAG_TTH_LEAVES) == false) {
		char* buf = new char[STRING(FINISHED_UPLOAD).size() + MAX_PATH + 128];
		sprintf(buf, CSTRING(FINISHED_UPLOAD), (Util::getFileName(u->getLocalFileName())).c_str(), 
			u->getUserConnection()->getUser()->getNick().c_str());
		LogManager::getInstance()->message(buf, false);
		delete[] buf;
	}
}

/* Some parts Copyright (C) 2002-2005 Petr Kozelka, PPK at seznam dot cz */
