/* 
 * Copyright (C) 2001-2005 Jacek Sieka, arnetheduck on gmail point com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "stdafx.h"
#include "../client/DCPlusPlus.h"
#include "Resource.h"

#include "HubFrame.h"
#include "LineDlg.h"
#include "SearchFrm.h"
#include "PrivateFrame.h"

#include "../client/QueueManager.h"
#include "../client/ShareManager.h"
#include "../client/Util.h"
#include "../client/FavoriteManager.h"
#include "../client/LogManager.h"
#include "../client/SettingsManager.h"
#include "../client/ConnectionManager.h" 

HubFrame::FrameMap HubFrame::frames;

int HubFrame::columnSizes[] = { 100, 75, 75, 75, 100, 75, 100, 40, 40, 40, 40, 40, 40, 50 };
int HubFrame::columnIndexes[] = { COLUMN_NICK, COLUMN_SHARED, COLUMN_EXACT_SHARED, COLUMN_DESCRIPTION, COLUMN_TAG, 
	COLUMN_CONNECTION, COLUMN_EMAIL, COLUMN_CLIENTID, COLUMN_VERSION, COLUMN_MODE, COLUMN_HUBS, COLUMN_SLOTS, 
	COLUMN_LIMITER, COLUMN_IP };
static ResourceManager::Strings columnNames[] = { ResourceManager::NICK, ResourceManager::SHARED, ResourceManager::EXACT_SHARED, 
ResourceManager::DESCRIPTION, ResourceManager::TAG, ResourceManager::CONNECTION, ResourceManager::EMAIL, 
ResourceManager::CLIENTID, ResourceManager::VERSION, ResourceManager::MODE, ResourceManager::HUBS, ResourceManager::SLOTS, 
ResourceManager::LIMIT, ResourceManager::SETTINGS_IP };

tstring sSelectedURL = Util::emptyStringT;
long lURLBegin = 0;
long lURLEnd = 0;

LRESULT HubFrame::OnCreate(UINT /*uMsg*/, WPARAM /*wParam*/, LPARAM /*lParam*/, BOOL& bHandled)
{
	CreateSimpleStatusBar(ATL_IDS_IDLEMESSAGE, WS_CHILD | WS_VISIBLE | WS_CLIPCHILDREN | WS_CLIPSIBLINGS | SBARS_SIZEGRIP);
	ctrlStatus.Attach(m_hWndStatusBar);

	ctrlClient.Create(m_hWnd, rcDefault, NULL, WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | 
		WS_VSCROLL | ES_MULTILINE | ES_NOHIDESEL | ES_READONLY, WS_EX_CLIENTEDGE, IDC_CLIENT);

	ctrlClient.LimitText(0);
	ctrlClient.SetFont(WinUtil::font);
	clientContainer.SubclassWindow(ctrlClient.m_hWnd);
	
	ctrlClient.SetAutoURLDetect(false);
	ctrlClient.SetEventMask( ctrlClient.GetEventMask() | ENM_LINK );
	ctrlClient.SetUsers(&ctrlUsers);
	
	ctrlMessage.Create(m_hWnd, rcDefault, NULL, WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | 
		ES_AUTOHSCROLL | ES_AUTOVSCROLL | ES_MULTILINE, WS_EX_CLIENTEDGE);
	
	ctrlMessageContainer.SubclassWindow(ctrlMessage.m_hWnd);
	ctrlMessage.SetFont(WinUtil::font);
	ctrlMessage.SetLimitText(9999);

	ctrlFilter.Create(m_hWnd, rcDefault, NULL, WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | 
		ES_AUTOHSCROLL | WS_BORDER, WS_EX_CLIENTEDGE);

	ctrlFilterContainer.SubclassWindow(ctrlFilter.m_hWnd);
	ctrlFilter.SetFont(WinUtil::font);

	ctrlFilterSel.Create(m_hWnd, rcDefault, NULL, WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_HSCROLL |
		WS_VSCROLL | CBS_DROPDOWNLIST | WS_BORDER, WS_EX_CLIENTEDGE);

	ctrlFilterSelContainer.SubclassWindow(ctrlFilterSel.m_hWnd);
	ctrlFilterSel.SetFont(WinUtil::font); 

	ctrlUsers.Create(m_hWnd, rcDefault, NULL, WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | 
		WS_HSCROLL | WS_VSCROLL | LVS_REPORT | LVS_SHOWSELALWAYS | LVS_SHAREIMAGELISTS, WS_EX_CLIENTEDGE, IDC_USERS);

	if(Util::isXP) {
		ctrlUsers.SetExtendedListViewStyle(LVS_EX_LABELTIP | LVS_EX_HEADERDRAGDROP | LVS_EX_FULLROWSELECT);
	} else {
		ctrlUsers.SetExtendedListViewStyle(LVS_EX_LABELTIP | LVS_EX_HEADERDRAGDROP | LVS_EX_FULLROWSELECT | 0x00010000);
	}

	splitChat.Create(m_hWnd);
	splitChat.SetSplitterPanes(ctrlClient.m_hWnd, ctrlMessage.m_hWnd, true);
	splitChat.SetSplitterExtendedStyle(0);
	splitChat.SetSplitterPos((WinUtil::fontHeight + 4)+5);

	SetSplitterPanes(splitChat.m_hWnd, ctrlUsers.m_hWnd, false);
	SetSplitterExtendedStyle(SPLIT_PROPORTIONAL);
	if(hubchatusersplit){ m_nProportionalPos = hubchatusersplit;
	} else { m_nProportionalPos = 7500; }
	ctrlShowUsers.Create(ctrlStatus.m_hWnd, rcDefault, _T("+/-"), WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS | WS_CLIPCHILDREN);
	ctrlShowUsers.SetButtonStyle(BS_AUTOCHECKBOX, false);
	ctrlShowUsers.SetFont(WinUtil::systemFont);
	ctrlShowUsers.SetCheck(ShowUserList ? BST_CHECKED : BST_UNCHECKED);
	showUsersContainer.SubclassWindow(ctrlShowUsers.m_hWnd);

    bool bColumsFromFavorite = false;
    if(sColumsOrder != Util::emptyString && sColumsWidth != Util::emptyString && sColumsVisible != Util::emptyString) bColumsFromFavorite = true;
    if(bColumsFromFavorite == false) {
		WinUtil::splitTokens(columnIndexes, SETTING(HUBFRAME_ORDER), COLUMN_LAST);
		WinUtil::splitTokens(columnSizes, SETTING(HUBFRAME_WIDTHS), COLUMN_LAST);                           
    } else {
		WinUtil::splitTokens(columnIndexes, sColumsOrder, COLUMN_LAST);
		WinUtil::splitTokens(columnSizes, sColumsWidth, COLUMN_LAST);
	}
    	
	for(int j=0; j<COLUMN_LAST; j++) {
		int fmt = (j == COLUMN_SHARED || j == COLUMN_EXACT_SHARED) ? LVCFMT_RIGHT : LVCFMT_LEFT;
		ctrlUsers.InsertColumn(j, CTSTRING_I(columnNames[j]), fmt, columnSizes[j], j);
	}
	
	ctrlUsers.setColumnOrderArray(COLUMN_LAST, columnIndexes);
    if(bColumsFromFavorite == true) {
	    ctrlUsers.setVisible(sColumsVisible);
    } else {
	    ctrlUsers.setVisible(SETTING(HUBFRAME_VISIBLE));
    }

	ctrlUsers.SetBkColor(WinUtil::bgColor);
	ctrlUsers.SetTextBkColor(WinUtil::bgColor);
	ctrlUsers.SetTextColor(WinUtil::textColor);
	ctrlUsers.setOwnerDraw(true);
	ctrlClient.SetBackgroundColor(WinUtil::bgColor); 
	
	ctrlUsers.setSortColumn(UserInfo::COLUMN_NICK);
				
	ctrlUsers.SetImageList(WinUtil::userImages, LVSIL_SMALL);

	CToolInfo ti(TTF_SUBCLASS, ctrlStatus.m_hWnd);
	
	ctrlLastLines.Create(ctrlStatus.m_hWnd, rcDefault, NULL, WS_POPUP | TTS_NOPREFIX | TTS_ALWAYSTIP, WS_EX_TOPMOST);
	ctrlLastLines.SetWindowPos(HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);
	ctrlLastLines.AddTool(&ti);

	tabcopyMenu.CreatePopupMenu();
	tabcopyMenu.AppendMenu(MF_STRING, IDC_COPY_HUB, CTSTRING(HUB_NAME));
	tabcopyMenu.AppendMenu(MF_STRING, IDC_COPY_LINK, CTSTRING(HUB_ADDRESS));

	tabMenu.CreatePopupMenu();
	if(BOOLSETTING(LOG_MAIN_CHAT)) {
		tabMenu.AppendMenu(MF_STRING, IDC_OPEN_HUB_LOG, CTSTRING(OPEN_HUB_LOG));
		tabMenu.AppendMenu(MF_SEPARATOR);
	}
	tabMenu.AppendMenu(MF_STRING, IDC_ADD_AS_FAVORITE, CTSTRING(ADD_TO_FAVORITES));
	tabMenu.AppendMenu(MF_STRING, ID_FILE_RECONNECT, CTSTRING(MENU_RECONNECT));
	tabMenu.AppendMenu(MF_POPUP, (UINT)(HMENU)tabcopyMenu, CTSTRING(COPY));

	showJoins = BOOLSETTING(SHOW_JOINS);
	favShowJoins = BOOLSETTING(FAV_SHOW_JOINS);

	ctrlFilterSel.AddString(CTSTRING(NICK));
	ctrlFilterSel.AddString(CTSTRING(SHARED));
	ctrlFilterSel.AddString(CTSTRING(EXACT_SHARED));
	ctrlFilterSel.AddString(CTSTRING(DESCRIPTION));
	ctrlFilterSel.AddString(CTSTRING(TAG));
	ctrlFilterSel.AddString(CTSTRING(CONNECTION));
	ctrlFilterSel.AddString(CTSTRING(EMAIL));
	ctrlFilterSel.AddString(CTSTRING(CLIENTID));
	ctrlFilterSel.AddString(CTSTRING(VERSION));
	ctrlFilterSel.AddString(CTSTRING(MODE));
	ctrlFilterSel.AddString(CTSTRING(HUBS));
	ctrlFilterSel.AddString(CTSTRING(SLOTS));
	ctrlFilterSel.AddString(CTSTRING(LIMIT));
	ctrlFilterSel.AddString(CTSTRING(SETTINGS_IP));
	ctrlFilterSel.SetCurSel(0);

	bHandled = FALSE;
	client->connect();
	
	TimerManager::getInstance()->addListener(this);
	SettingsManager::getInstance()->addListener(this);
	
	return 1;
}

void HubFrame::openWindow(const tstring& aServer, 
		int windowposx, int windowposy, int windowsizex, int windowsizey, int windowtype, int chatusersplit, bool userliststate, 
        string sColumsOrder, string sColumsWidth, string sColumsVisible) {
	FrameIter i = frames.find(aServer);
	if(i == frames.end()) {
		HubFrame* frm = new HubFrame(aServer, 
			windowposx, windowposy, windowsizex, windowsizey, windowtype, chatusersplit, userliststate, 
            sColumsOrder, sColumsWidth, sColumsVisible);
		frames[aServer] = frm;

		int nCmdShow = SW_SHOWDEFAULT;
		CRect rc = frm->rcDefault;

		rc.left = windowposx;
		rc.top = windowposy;
		rc.right = rc.left + windowsizex;
		rc.bottom = rc.top + windowsizey;
		if( (rc.left < 0 ) || (rc.top < 0) || (rc.right - rc.left < 10) || ((rc.bottom - rc.top) < 10) ) {
			rc = frm->rcDefault;
		}
		frm->CreateEx(WinUtil::mdiClient, rc);
		if(windowtype)
			frm->ShowWindow(((nCmdShow == SW_SHOWDEFAULT) || (nCmdShow == SW_SHOWNORMAL)) ? windowtype : nCmdShow);
	} else {
		if(::IsIconic(i->second->m_hWnd))
			::ShowWindow(i->second->m_hWnd, SW_RESTORE);
		i->second->MDIActivate(i->second->m_hWnd);
	}
}

void HubFrame::onEnter() {
	if(ctrlMessage.GetWindowTextLength() > 0) {
		AutoArray<TCHAR> msg(ctrlMessage.GetWindowTextLength()+1);
		ctrlMessage.GetWindowText(msg, ctrlMessage.GetWindowTextLength()+1);
		tstring s(msg, ctrlMessage.GetWindowTextLength());

		// save command in history, reset current buffer pointer to the newest command
		curCommandPosition = prevCommands.size();		//this places it one position beyond a legal subscript
		if (!curCommandPosition || curCommandPosition > 0 && prevCommands[curCommandPosition - 1] != s) {
			++curCommandPosition;
			prevCommands.push_back(s);
		}
		currentCommand = _T("");

		// Special command
		if(s[0] == _T('/')) {
			tstring m = s;
			tstring param;
			tstring message;
			tstring status;

			if(BOOLSETTING(CLIENT_COMMANDS)) {
				if(!s.empty())
					addClientLine(_T("Client command: ") + s, WinUtil::m_ChatTextSystem);
			}

			if(WinUtil::checkCommand(s, param, message, status)) {
				if(!message.empty()) {
					client->hubMessage(Text::fromT(message));
				}
				if(!status.empty()) {
					addClientLine(status, WinUtil::m_ChatTextSystem);
				}
			} else if(Util::stricmp(s.c_str(), _T("join"))==0) {
				if(!param.empty()) {
					redirect = param;
					if(BOOLSETTING(SETTINGS_OPEN_NEW_WINDOW)) {
						HubFrame::openWindow(param);
					} else {
						BOOL whatever = FALSE;
						onFollow(0, 0, 0, whatever);
					}
				} else {
					addClientLine(TSTRING(SPECIFY_SERVER), WinUtil::m_ChatTextSystem);
				}
			} else if((Util::stricmp(s.c_str(), _T("clear")) == 0) || (Util::stricmp(s.c_str(), _T("cls")) == 0)) {
				ctrlClient.SetWindowText(_T(""));
			} else if(Util::stricmp(s.c_str(), _T("ts")) == 0) {
				timeStamps = !timeStamps;
				if(timeStamps) {
					addClientLine(TSTRING(TIMESTAMPS_ENABLED), WinUtil::m_ChatTextSystem);
				} else {
					addClientLine(TSTRING(TIMESTAMPS_DISABLED), WinUtil::m_ChatTextSystem);
				}
			} else if( (Util::stricmp(s.c_str(), _T("password")) == 0) && waitingForPW ) {
				client->setPassword(Text::fromT(param));
				client->password(Text::fromT(param));
				waitingForPW = false;
			} else if( Util::stricmp(s.c_str(), _T("showjoins")) == 0 ) {
				showJoins = !showJoins;
				if(showJoins) {
					addClientLine(TSTRING(JOIN_SHOWING_ON), WinUtil::m_ChatTextSystem);
				} else {
					addClientLine(TSTRING(JOIN_SHOWING_OFF), WinUtil::m_ChatTextSystem);
				}
			} else if( Util::stricmp(s.c_str(), _T("favshowjoins")) == 0 ) {
				favShowJoins = !favShowJoins;
				if(favShowJoins) {
					addClientLine(TSTRING(FAV_JOIN_SHOWING_ON), WinUtil::m_ChatTextSystem);
				} else {
					addClientLine(TSTRING(FAV_JOIN_SHOWING_OFF), WinUtil::m_ChatTextSystem);
				}
			} else if(Util::stricmp(s.c_str(), _T("close")) == 0) {
				PostMessage(WM_CLOSE);
			} else if(Util::stricmp(s.c_str(), _T("userlist")) == 0) {
				ctrlShowUsers.SetCheck(ShowUserList ? BST_UNCHECKED : BST_CHECKED);
			} else if(Util::stricmp(s.c_str(), _T("connection")) == 0) {
				addClientLine(Text::toT((STRING(IP) + client->getLocalIp() + ", " + STRING(PORT) + Util::toString(SETTING(IN_PORT)) + "/" + Util::toString(SETTING(UDP_PORT)))), WinUtil::m_ChatTextSystem);
			} else if((Util::stricmp(s.c_str(), _T("favorite")) == 0) || (Util::stricmp(s.c_str(), _T("fav")) == 0)) {
				addAsFavorite();
			} else if(Util::stricmp(s.c_str(), _T("getlist")) == 0){
				if( !param.empty() ){
					int k = ctrlUsers.findItem(param);
					if(k != -1) {
						ctrlUsers.getItemData(k)->getList();
					}
				}
			} else if(Util::stricmp(s.c_str(), _T("extraslots"))==0) {
				int j = Util::toInt(Text::fromT(param));
				if(j > 0) {
					SettingsManager::getInstance()->set(SettingsManager::EXTRA_SLOTS, j);
					addClientLine(TSTRING(EXTRA_SLOTS_SET), WinUtil::m_ChatTextSystem );
				} else {
					addClientLine(TSTRING(INVALID_NUMBER_OF_SLOTS), WinUtil::m_ChatTextSystem );
				}
			} else if(Util::stricmp(s.c_str(), _T("smallfilesize"))==0) {
				int j = Util::toInt(Text::fromT(param));
				if(j >= 64) {
					SettingsManager::getInstance()->set(SettingsManager::SMALL_FILE_SIZE, j);
					addClientLine(TSTRING(SMALL_FILE_SIZE_SET), WinUtil::m_ChatTextSystem );
				} else {
					addClientLine(TSTRING(INVALID_SIZE), WinUtil::m_ChatTextSystem );
				}
			} else if(Util::stricmp(s.c_str(), _T("savequeue")) == 0) {
				QueueManager::getInstance()->saveQueue();
				addClientLine(_T("Queue saved."), WinUtil::m_ChatTextSystem );
			} else if(Util::stricmp(s.c_str(), _T("whois")) == 0) {
				WinUtil::openLink(_T("http://www.ripe.net/perl/whois?form_type=simple&full_query_string=&searchtext=") + Text::toT(Util::encodeURI(Text::fromT(param))));
			} else if(Util::stricmp(s.c_str(), _T("ignorelist"))==0) {
				tstring ignorelist = _T("Ignored users:");
				for(TStringHash::iterator i = ignoreList.begin(); i != ignoreList.end(); ++i)
					ignorelist += _T(" ") + *i;
				addLine(ignorelist, WinUtil::m_ChatTextSystem);
			} else if(Util::stricmp(s.c_str(), _T("log")) == 0) {
				StringMap params;
				params["hub"] = client->getName();
				params["hubaddr"] = client->getAddressPort();
				params["mynick"] = client->getNick(); 
				if(param.empty()) {
					WinUtil::openFile(Text::toT(Util::validateFileName(SETTING(LOG_DIRECTORY) + Util::formatParams(SETTING(LOG_FILE_MAIN_CHAT), params))));
				} else if(Util::stricmp(param.c_str(), _T("status")) == 0) {
					WinUtil::openFile(Text::toT(Util::validateFileName(SETTING(LOG_DIRECTORY) + Util::formatParams(SETTING(LOG_FILE_STATUS), params))));
				}
			} else if(Util::stricmp(s.c_str(), _T("help")) == 0) {
				addLine(_T("*** ") + WinUtil::commands + _T(", /smallfilesize #, /extraslots #, /savequeue, /join <hub-ip>, /clear, /ts, /showjoins, /favshowjoins, /close, /userlist, /connection, /favorite, /pm <user> [message], /getlist <user>, /log <status, system, downloads, uploads>, /winamp, /showblockedipports, /whois [IP], /ignorelist, /favoritenick, /favoritepassword"), WinUtil::m_ChatTextSystem);
			} else if(Util::stricmp(s.c_str(), _T("pm")) == 0) {
				string::size_type j = param.find(_T(' '));
				if(j != string::npos) {
					tstring nick = param.substr(0, j);
					int k = ctrlUsers.findItem(nick);
					if(k != -1) {
						UserInfo* ui = ctrlUsers.getItemData(k);
						if(param.size() > j + 1)
							PrivateFrame::openWindow(ui->user, param.substr(j+1));
						else
							PrivateFrame::openWindow(ui->user);
					}
				} else if(!param.empty()) {
					int k = ctrlUsers.findItem(param);
					if(k != -1) {
						UserInfo* ui = ctrlUsers.getItemData(k);
						PrivateFrame::openWindow(ui->user);
					}
				}
			} else if(Util::stricmp(s.c_str(), _T("me")) == 0) {
				client->hubMessage(Text::fromT(m));
			} else if((Util::stricmp(s.c_str(), _T("favoritenick")) == 0) || (Util::stricmp(s.c_str(), _T("favnick")) == 0)) {
				FavoriteHubEntry* hub = FavoriteManager::getInstance()->getFavoriteHubEntry(Text::fromT(server));
				if(hub) {
					if(!param.empty()) {
						hub->setNick(Text::fromT(param));
					} else {
						hub->setNick(client->getNick());
					}
					FavoriteManager::getInstance()->save();
					addClientLine(TSTRING(NICK_FOR_FAVORITE), WinUtil::m_ChatTextSystem );
				} else {
					addClientLine(TSTRING(HUB_NOT_FAVORITE), WinUtil::m_ChatTextSystem );
				}
			} else if((Util::stricmp(s.c_str(), _T("favoritepassword")) == 0) || (Util::stricmp(s.c_str(), _T("favpass")) == 0)) {
				FavoriteHubEntry* hub = FavoriteManager::getInstance()->getFavoriteHubEntry(Text::fromT(server));
				if(hub) {
					if(!param.empty()) {
						hub->setPassword(Text::fromT(param));
					} else {
						hub->setPassword(client->getPassword());
					}
					FavoriteManager::getInstance()->save();
					addClientLine(TSTRING(PASS_FOR_FAVORITE), WinUtil::m_ChatTextSystem );
				} else {
					addClientLine(TSTRING(HUB_NOT_FAVORITE), WinUtil::m_ChatTextSystem );
				}
			} else if((Util::stricmp(s.c_str(), _T("stat")) == 0) || (Util::stricmp(s.c_str(), _T("stats")) == 0)) {
			    if(client->getOp())
					client->hubMessage(WinUtil::generateStats());
			    else
					addLine(Text::toT(WinUtil::generateStats()));
			} else {
				if (BOOLSETTING(SEND_UNKNOWN_COMMANDS)) {
					client->hubMessage(Text::fromT(m));
				} else {
					addClientLine(TSTRING(UNKNOWN_COMMAND) + s);
				}
			}
			ctrlMessage.SetWindowText(_T(""));
		} else if(waitingForPW) {
			addClientLine(TSTRING(DONT_REMOVE_SLASH_PASSWORD));
			ctrlMessage.SetWindowText(_T("/password "));
			ctrlMessage.SetFocus();
			ctrlMessage.SetSel(10, 10);
		} else {
			if(BOOLSETTING(SERVER_COMMANDS)) {
				if(((s[0] == '!') || (s[0] == '+') || (s[0] == '-')) && !s.empty())
					addClientLine(_T("Server command: ") + s, WinUtil::m_ChatTextSystem);
			}

			client->hubMessage(Text::fromT(s));
			ctrlMessage.SetWindowText(_T(""));
		}
	} else {
		MessageBeep(MB_ICONEXCLAMATION);
	}
}

struct CompareItems {
	CompareItems(int aCol) : col(aCol) { }
	bool operator()(const UserInfo& a, const UserInfo& b) const {
		return UserInfo::compareItems(&a, &b, col) < 0;
	}
	const int col;
};

int HubFrame::findUser(const User::Ptr& aUser) {
	UserMapIter i = userMap.find(aUser);
	if(i == userMap.end())
		return -1;

	if(ctrlUsers.getSortColumn() == UserInfo::COLUMN_NICK) {
		// Sort order of the other columns changes too late when the user's updated
		UserInfo* ui = i->second;
		for(int j = 0; j < ctrlUsers.GetItemCount(); ++j) {
			if(ctrlUsers.getItemData(j) == ui)
				return j;
		}
	}
	return ctrlUsers.findItem(Text::toT(aUser->getNick()));
}

void HubFrame::addAsFavorite() {
	FavoriteHubEntry aEntry;
	TCHAR buf[256];
	this->GetWindowText(buf, 255);
	aEntry.setServer(Text::fromT(server));
	aEntry.setName(Text::fromT(buf));
	aEntry.setDescription(Text::fromT(buf));
	aEntry.setConnect(false);
	aEntry.setNick(client->getNick());
	aEntry.setPassword(client->getPassword());
	aEntry.setConnect(false);
	FavoriteManager::getInstance()->addFavorite(aEntry);
	addClientLine(TSTRING(FAVORITE_HUB_ADDED), WinUtil::m_ChatTextSystem );
}

LRESULT HubFrame::onCopyUserInfo(WORD /*wNotifyCode*/, WORD wID, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
    if(client->isConnected()) {
        string sCopy;
		UserInfo* ui;

		if(sSelectedUser != _T("")) {
			int i = ctrlUsers.findItem(sSelectedUser);
			if ( i >= 0 ) {
				ui = (UserInfo*)ctrlUsers.getItemData(i);

				switch (wID) {
					case IDC_COPY_NICK:
						sCopy += ui->user->getNick();
						break;
					case IDC_COPY_EXACT_SHARE:
						sCopy += Util::formatExactSize(ui->user->getBytesShared());
						break;
					case IDC_COPY_DESCRIPTION:
						sCopy += ui->user->getDescription();
						break;
					case IDC_COPY_TAG:
						sCopy += ui->user->getTag();
						break;
					case IDC_COPY_EMAIL_ADDRESS:
						sCopy += ui->user->getEmail();
						break;
					case IDC_COPY_IP:
						sCopy += ui->user->getIp();
						break;
					default:
						dcdebug("HUBFRAME DON'T GO HERE\n");
						return 0;
				}
			}
		}
		if(!sCopy.empty())
			WinUtil::setClipboard(Text::toT(sCopy));
    }
	return 0;
}

LRESULT HubFrame::onDoubleClickUsers(int /*idCtrl*/, LPNMHDR pnmh, BOOL& /*bHandled*/) {
	NMITEMACTIVATE* item = (NMITEMACTIVATE*)pnmh;
	if(client->isConnected() && item->iItem != -1) {
	    switch(SETTING(USERLIST_DBLCLICK)) {
		    case 0:
		          ctrlUsers.getItemData(item->iItem)->getList();
		          break;
		    case 1: {
	              CAtlString sUser = Text::toT(ctrlUsers.getItemData(item->iItem)->user->getNick()).c_str();
	              CAtlString sText = "";
	              int iSelBegin, iSelEnd;
	              ctrlMessage.GetSel(iSelBegin, iSelEnd);
	              ctrlMessage.GetWindowText(sText);

	              if((iSelBegin == 0) && (iSelEnd == 0)) {
                        if(sText.GetLength() == 0) {
			                  sUser += ": ";
			                  ctrlMessage.SetWindowText(sUser);
			                  ctrlMessage.SetFocus();
			                  ctrlMessage.SetSel(ctrlMessage.GetWindowTextLength(), ctrlMessage.GetWindowTextLength());
                        } else {
			                  sUser += ": ";
			                  ctrlMessage.ReplaceSel(sUser);
			                  ctrlMessage.SetFocus();
                        }
	              } else {
                        sUser += " ";
                        ctrlMessage.ReplaceSel(sUser);
                        ctrlMessage.SetFocus();
	              }
	              break;
		    }    
		    case 2:
		          if(ctrlUsers.getItemData(item->iItem)->user->getNick() != sMyNick)
                        ctrlUsers.getItemData(item->iItem)->pm();
		          break;
		    case 3:
		          ctrlUsers.getItemData(item->iItem)->matchQueue();
		          break;
		    case 4:
		          ctrlUsers.getItemData(item->iItem)->grant();
		          break;
		    case 5:
		          ctrlUsers.getItemData(item->iItem)->addFav();
		          break;
		}	
	}
	return 0;
}

LRESULT HubFrame::onEditCopy(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
	HWND focus = GetFocus();
	if ( focus == ctrlClient.m_hWnd ) {
		ctrlClient.Copy();
	}
	return 0;
}

LRESULT HubFrame::onEditSelectAll(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
	HWND focus = GetFocus();
	if ( focus == ctrlClient.m_hWnd ) {
		ctrlClient.SetSelAll();
	}
	return 0;
}

LRESULT HubFrame::onEditClearAll(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
	HWND focus = GetFocus();
	if ( focus == ctrlClient.m_hWnd ) {
		ctrlClient.SetWindowText(_T(""));
	}
	return 0;
}

bool HubFrame::updateUser(const User::Ptr& u, bool searchinlist /* = true */) {
	int i = -1;
	if(searchinlist)
		i = findUser(u);
		
	if(i == -1) {
		if(filterstring.empty()) {
			UserInfo* ui = new UserInfo(u);
			userMap.insert(make_pair(u, ui));
			ctrlUsers.insertItem(ui, WinUtil::getImage(u));
			return true;
		} else {
			UserMapIter j = userMap.find(u);
			if(j == userMap.end()) {
				bool adduser = false;
				if(filterstring.empty()) {
					adduser = true;
				} else if(filterUser(u) == true) {
					adduser = true;
				}
				UserInfo* ui = new UserInfo(u);
				userMap.insert(make_pair(u, ui));
				if(adduser == true) {
					ctrlUsers.insertItem(ui, WinUtil::getImage(u));
				}
				return true;
			} else if(filterUser(u) == true) {
				ctrlUsers.insertItem(j->second, WinUtil::getImage(j->second->user));
				return false;
			}
			return false;
		}
	} else {
		bool deleteuser = true;
		if(filterstring.empty()) {
			deleteuser = false;
		} else if(filterUser(u) == true) {
			deleteuser = false;
		}
		if(deleteuser == false) {
			UserInfo* ui = ctrlUsers.getItemData(i);
			bool resort = (ui->getOp() != u->isSet(User::OP));
			ctrlUsers.getItemData(i)->update();
			ctrlUsers.updateItem(i);
			ctrlUsers.SetItem(i, 0, LVIF_IMAGE, NULL, WinUtil::getImage(u), 0, 0, NULL);
			if(resort)
				ctrlUsers.resort();
		} else {
			ctrlUsers.SetItemState(i, 0, LVIS_SELECTED);
			ctrlUsers.DeleteItem(i);
		}
		return false;
	}
}

static const char* sSameNumbers[] = { "000000", "111111", "222222", "333333", "444444", "555555", "666666", "777777", "888888", "999999" };

LRESULT HubFrame::onSpeaker(UINT /*uMsg*/, WPARAM wParam, LPARAM lParam, BOOL& /*bHandled*/) {
	if(wParam == UPDATE_USERS) {
		ctrlUsers.SetRedraw(FALSE);
		bool resort = false;
		{
			Lock l(updateCS);
			for(UpdateIter i = updateList.begin(); i != updateList.end(); ++i) {
				User::Ptr& u = i->first;
				switch(i->second) {
				case UPDATE_USER:
					if(updateUser(u)) {
						if(showJoins || (favShowJoins && u->isFavoriteUser())) {
						 	addLine(_T("*** ") + TSTRING(JOINS) + Text::toT(u->getNick()), WinUtil::m_ChatTextSystem);
						}
						if(client->getOp() && BOOLSETTING(SHOW_SHARE_CHEAT)) {
							int64_t bytesSharedInt64 = u->getBytesShared();
							if(bytesSharedInt64 > 0) {
								string bytesShared = Util::toString(bytesSharedInt64);
								bool divided = (bytesSharedInt64 % (1024*1024) == 0);
								bool samenumbers = false;
								for(int i = 0; i < 10; i++) {
									if(strstr(bytesShared.c_str(), sSameNumbers[i]) != 0) {
										samenumbers = true;
										break;
									}
								}
								char buf[512];
								if(samenumbers && divided) {
									sprintf(buf, "*** User %s %s %s bytes", u->getNick().c_str(), (STRING(SAMENUMBERS_DIVIDED)).c_str(), bytesShared.c_str());
								} else if(samenumbers) {
									sprintf(buf, "*** User %s %s %s bytes", u->getNick().c_str(), (STRING(MANY_SAMENUMBERS)).c_str(), bytesShared.c_str());
								} else if(divided) {
									sprintf(buf, "*** User %s %s %s bytes", u->getNick().c_str(), (STRING(DIVIDED_MB)).c_str(), bytesShared.c_str());
								}
								if(samenumbers || divided) {
									addLine(Text::toT(buf), WinUtil::m_ChatTextSystem);
								}
							}
						}
					} else {
						resort = true;
					}
					break;
				case UPDATE_USERS:
					if(!updateUser(u))
						resort = true;
					
					break;
				case REMOVE_USER:
					int j = findUser(u);
					if( j != -1 ) {
						UserInfo* ui = ctrlUsers.getItemData(j); 
						ctrlUsers.SetItemState(j, 0, LVIS_SELECTED);
						ctrlUsers.DeleteItem(j);
						delete ui;
						if(showJoins || (favShowJoins && u->isFavoriteUser())) {
							addLine(Text::toT("*** " + STRING(PARTS) + u->getNick()), WinUtil::m_ChatTextSystem);
						}
						userMap.erase(u);
					} else {
						UserMapIter i = userMap.find(u);
						if(i != userMap.end()) {
							delete i->second;
							userMap.erase(u);
						}
					}
					break;
				}
		}
			updateList.clear();
		}
		if(ctrlUsers.getSortColumn() != UserInfo::COLUMN_NICK)
			ctrlUsers.resort();
		ctrlUsers.SetRedraw(TRUE);
	} else if(wParam == DISCONNECTED) {
		clearUserList();
		setTabColor(RGB(255, 0, 0));
		setIconState();
	} else if(wParam == CONNECTED) {
		addClientLine(TSTRING(CONNECTED), WinUtil::m_ChatTextServer);
		setTabColor(RGB(0, 255, 0));
		unsetIconState();
	} else if(wParam == ADD_CHAT_LINE) {
		tstring* x = (tstring*)lParam;
		int nickPos = x->find('<') + 1;
		tstring nick = x->substr(nickPos, x->find('>') - nickPos);
		int i = -1;
		if(nick != _T(""))
			i = ctrlUsers.findItem(nick);
		if(i >= 0) {
			UserInfo* ui = (UserInfo*)ctrlUsers.getItemData(i);
			if (!ignoreList.count(nick) || (ui->user->isSet(User::OP) && !client->getOp()))
				addLine(*x, WinUtil::m_ChatTextGeneral);
		} else {
			if (!ignoreList.count(nick))
				addLine(*x, WinUtil::m_ChatTextGeneral);
		}
		delete x;
	} else if(wParam == ADD_STATUS_LINE) {
		tstring* x = (tstring*)lParam;
		addClientLine(*x, WinUtil::m_ChatTextServer );
		delete x;
	} else if(wParam == ADD_SILENT_STATUS_LINE) {
		tstring* x = (tstring*)lParam;
		addClientLine(*x, false);
		delete x;
	} else if(wParam == SET_WINDOW_TITLE) {
		tstring* x = (tstring*)lParam;
		SetWindowText(x->c_str());
		delete x;
	} else if(wParam == STATS) {
		size_t AllUsers = client->getUserCount();
		size_t ShownUsers = ctrlUsers.GetItemCount();
		if(filterstring.empty() == false && AllUsers != ShownUsers) {
			ctrlStatus.SetText(1, Text::toT(Util::toString(ShownUsers) + "/" + Util::toString(AllUsers) + " " + STRING(HUB_USERS)).c_str());
		} else {
			ctrlStatus.SetText(1, Text::toT(Util::toString(client->getUserCount()) + " " + STRING(HUB_USERS)).c_str());
		}
		ctrlStatus.SetText(2, Text::toT(Util::formatBytes(client->getAvailable())).c_str());
	} else if(wParam == GET_PASSWORD) {
		if(client->getPassword().size() > 0) {
			client->password(client->getPassword());
			addClientLine(TSTRING(STORED_PASSWORD_SENT), WinUtil::m_ChatTextSystem);
		} else {
			ctrlMessage.SetWindowText(_T("/password "));
			ctrlMessage.SetFocus();
			ctrlMessage.SetSel(10, 10);
			waitingForPW = true;
		}
	} else if(wParam == PRIVATE_MESSAGE) {
		PMInfo* i = (PMInfo*)lParam;
		if(!ignoreList.count(Text::toT(i->user->getNick())) || (i->user->isSet(User::OP) && !client->getOp())) {
			if(i->user->isOnline()) {
				if(BOOLSETTING(POPUP_PMS) || PrivateFrame::isOpen(i->user)) {
					PrivateFrame::gotMessage(i->user, i->msg);
				} else {
					addLine(TSTRING(PRIVATE_MESSAGE_FROM) + Text::toT(i->user->getNick()) + _T(": ") + i->msg, WinUtil::m_ChatTextPrivate);
				}
				if(BOOLSETTING(MINIMIZE_TRAY) && i->user->getConnection().empty() == false) {
					HWND hMainWnd = GetTopLevelWindow();
					::PostMessage(hMainWnd, WM_SPEAKER, 7, NULL);
				}
			} else {
				if(BOOLSETTING(IGNORE_OFFLINE)) {
					addClientLine(TSTRING(IGNORED_MESSAGE) + i->msg, WinUtil::m_ChatTextPrivate, false);
				} else if(BOOLSETTING(POPUP_OFFLINE)) {
					PrivateFrame::gotMessage(i->user, i->msg);
				} else {
					addLine(TSTRING(PRIVATE_MESSAGE_FROM) + Text::toT(i->user->getNick()) + _T(": ") + i->msg, WinUtil::m_ChatTextPrivate);
				}
			}
		}
		delete i;
	} else if(wParam == KICK_MSG) {
		tstring* x = (tstring*)lParam;
		addLine(*x, WinUtil::m_ChatTextServer, false);
		delete x;
	}

	return 0;
};

void HubFrame::UpdateLayout(BOOL bResizeBars /* = TRUE */) {
	RECT rect;
	GetClientRect(&rect);
	// position bars and offset their dimensions
	UpdateBarsPosition(rect, bResizeBars);
	
	if(ctrlStatus.IsWindow()) {
		CRect sr;
		int w[4];
		ctrlStatus.GetClientRect(sr);

		int tmp = (sr.Width()) > 332 ? 232 : ((sr.Width() > 132) ? sr.Width()-100 : 32);
		
		w[0] = sr.right - tmp;
		w[1] = w[0] + (tmp-32)/2;
		w[2] = w[0] + (tmp-32);
		w[3] = w[2] + 16;
		
		ctrlStatus.SetParts(4, w);

		ctrlLastLines.SetMaxTipWidth(w[0]);

		// Strange, can't get the correct width of the last field...
		ctrlStatus.GetRect(2, sr);
		sr.left = sr.right + 2;
		sr.right = sr.left + 16;
		ctrlShowUsers.MoveWindow(sr);
	}
	int h = WinUtil::fontHeight + 4;

	CRect rc = rect;
	if(!ShowUserList) {
		if(GetSinglePaneMode() == SPLIT_PANE_NONE)
			SetSinglePaneMode(SPLIT_PANE_LEFT);
	} else {
		if(GetSinglePaneMode() != SPLIT_PANE_NONE)
			SetSinglePaneMode(SPLIT_PANE_NONE);
	}
	SetSplitterRect(rc);

	if(rc.Height() > ((2 * h) + 5)) {
		splitChat.SetSplitterPos( rc.Height() - ((2 * h) + 5) );
	} else {
		splitChat.SetSplitterPos(rc.Height()-1);
	}

	if(ShowUserList){
		GetSplitterPaneRect(1, &rc);
		rc.bottom = rect.bottom - h - 10;
		ctrlUsers.MoveWindow(rc);
		
		rc.top = rc.bottom + 3;
		rc.bottom = rect.bottom - 2;
		rc.right = rc.right - 100;
		ctrlFilter.MoveWindow(rc);

		rc.left = rc.right + 4;
		rc.right = rect.right;
		rc.bottom = rc.bottom + 120;
		ctrlFilterSel.MoveWindow(rc); 
	}
}

LRESULT HubFrame::onClose(UINT /*uMsg*/, WPARAM /*wParam*/, LPARAM /*lParam*/, BOOL& bHandled) {
	if(!closed) {
		if(MainFrame::closing == true || !BOOLSETTING(CONFIRM_HUBEXIT) || (MessageBox(CTSTRING(CONFIRM_HUBCLOSE), _T(APPNAME) _T(" ") _T(VERSIONSTRING), MB_YESNO | MB_ICONQUESTION | MB_DEFBUTTON2) == IDYES)) {
			RecentHubEntry* r = FavoriteManager::getInstance()->getRecentHubEntry(Text::fromT(server));
			if(r) {
				TCHAR buf[256];
				this->GetWindowText(buf, 255);
				r->setName(Text::fromT(buf));
				r->setUsers(Util::toString(client->getUserCount()));
				r->setShared(Util::toString(client->getAvailable()));
				FavoriteManager::getInstance()->updateRecent(r);
			}
			TimerManager::getInstance()->removeListener(this);
			SettingsManager::getInstance()->removeListener(this);
			client->removeListener(this);
			client->disconnect();
	
			closed = true;
			PostMessage(WM_CLOSE);
		}
		return 0;
	} else {
		SettingsManager::getInstance()->set(SettingsManager::GET_USER_INFO, ShowUserList);
		FavoriteManager::getInstance()->removeUserCommand(Text::fromT(server));

		clearUserList();
			
		FavoriteHubEntry* hub = FavoriteManager::getInstance()->getFavoriteHubEntry(Text::fromT(server));
		if(hub) {
			WINDOWPLACEMENT wp;
			wp.length = sizeof(wp);
			GetWindowPlacement(&wp);

			CRect rc;
			GetWindowRect(rc);
			CRect rcmdiClient;
			::GetWindowRect(WinUtil::mdiClient, &rcmdiClient);
			if(wp.showCmd == SW_SHOW || wp.showCmd == SW_SHOWNORMAL) {
				hub->setWindowPosX(rc.left - (rcmdiClient.left + 2));
				hub->setWindowPosY(rc.top - (rcmdiClient.top + 2));
				hub->setWindowSizeX(rc.Width());
				hub->setWindowSizeY(rc.Height());
			}
			if(wp.showCmd == SW_SHOWNORMAL || wp.showCmd == SW_SHOW || wp.showCmd == SW_SHOWMAXIMIZED || wp.showCmd == SW_MAXIMIZE)
				hub->setWindowType((int)wp.showCmd);
			hub->setChatUserSplit(m_nProportionalPos);
			hub->setUserListState(ShowUserList);
			ctrlUsers.saveFavoriteHeaderOrder(hub);
		} else {
			ctrlUsers.saveHeaderOrder(SettingsManager::HUBFRAME_ORDER, SettingsManager::HUBFRAME_WIDTHS,
				SettingsManager::HUBFRAME_VISIBLE);
		}
	
		bHandled = FALSE;
		return 0;
	}
}

LRESULT HubFrame::onLButton(UINT /*uMsg*/, WPARAM wParam, LPARAM lParam, BOOL& bHandled) {
	HWND focus = GetFocus();
	bHandled = false;
	if(focus == ctrlClient.m_hWnd) {
		POINT pt = { GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam) };
		tstring x;

		int i = ctrlClient.CharFromPos(pt);
		int line = ctrlClient.LineFromChar(i);
		int c = LOWORD(i) - ctrlClient.LineIndex(line);
		int len = ctrlClient.LineLength(i) + 1;
		if(len < 3) {
			return 0;
		}

		TCHAR* buf = new TCHAR[len];
		ctrlClient.GetLine(line, buf, len);
		x = tstring(buf, len-1);
		delete buf;

		string::size_type start = x.find_last_of(_T(" <\t\r\n"), c);
		if(start == string::npos)
			start = 0;
		else
			start++;

					
		string::size_type end = x.find_first_of(_T(" >\t"), start+1);

		if(end == string::npos) // get EOL as well
			end = x.length();
		else if(end == start + 1)
			return 0;

		// Nickname click, let's see if we can find one like it in the name list...
		int pos = ctrlUsers.findItem(x.substr(start, end - start));
		if(pos != -1) {
			bHandled = true;
			if (wParam & MK_CONTROL) { // MK_CONTROL = 0x0008
				PrivateFrame::openWindow((ctrlUsers.getItemData(pos))->user);
			} else if (wParam & MK_SHIFT) {
				try {
					QueueManager::getInstance()->addList((ctrlUsers.getItemData(pos))->user, QueueItem::FLAG_CLIENT_VIEW);
				} catch(const Exception& e) {
					addClientLine(Text::toT(e.getError()), WinUtil::m_ChatTextSystem);
				}
			} else {
				switch(SETTING(CHAT_DBLCLICK)) {
					case 0: {
					     int items = ctrlUsers.GetItemCount();
					     ctrlUsers.SetRedraw(FALSE);
					     for(int i = 0; i < items; ++i) {
					               ctrlUsers.SetItemState(i, (i == pos) ? LVIS_SELECTED | LVIS_FOCUSED : 0, LVIS_SELECTED | LVIS_FOCUSED);
					     }
					     ctrlUsers.SetRedraw(TRUE);
					     ctrlUsers.EnsureVisible(pos, FALSE);
					     break;
					}    
					case 1: {
					     CAtlString sUser = Text::toT(ctrlUsers.getItemData(pos)->user->getNick()).c_str();
					     CAtlString sText = "";
					     int iSelBegin, iSelEnd;
					     ctrlMessage.GetSel(iSelBegin, iSelEnd);
					     ctrlMessage.GetWindowText(sText);

					     if((iSelBegin == 0) && (iSelEnd == 0)) {
					          if(sText.GetLength() == 0) {
                                   sUser += ": ";
			                       ctrlMessage.SetWindowText(sUser);
			                       ctrlMessage.SetFocus();
			                       ctrlMessage.SetSel(ctrlMessage.GetWindowTextLength(), ctrlMessage.GetWindowTextLength());
					          } else {
			                       sUser += ": ";
			                       ctrlMessage.ReplaceSel(sUser);
			                       ctrlMessage.SetFocus();
					          }
					     } else {
					          sUser += " ";
					          ctrlMessage.ReplaceSel(sUser);
					          ctrlMessage.SetFocus();
					     }
					     break;
					}
					case 2:
					     if(ctrlUsers.getItemData(pos)->user->getNick() != sMyNick)
					          ctrlUsers.getItemData(pos)->pm();
					     break;
					case 3:
					     ctrlUsers.getItemData(pos)->getList();
					     break;
					case 4:
					     ctrlUsers.getItemData(pos)->matchQueue();
					     break;
					case 5:
					     ctrlUsers.getItemData(pos)->grant();
					     break;
					case 6:
					     ctrlUsers.getItemData(pos)->addFav();
					     break;
				}
			}
		}
	}
	return 0;
}

void HubFrame::addLine(const tstring& aLine) {
	addLine(aLine, WinUtil::m_ChatTextGeneral);
}

void HubFrame::addLine(const tstring& aLine, CHARFORMAT2& cf, bool bUseEmo/* = true*/) {
	ctrlClient.AdjustTextSize();

	tstring sTmp = aLine;
	tstring sAuthor = _T("");
	int iAuthorLen = 0;
	bool isMe = false;
	if(aLine[0] == _T('<')) {
		string::size_type i = aLine.find(_T(">"));
		if(i != string::npos) {
       		sAuthor = aLine.substr(1, i-1);
       		iAuthorLen = i;
			if(_tcsncmp(_T(" /me "), aLine.substr(i+1, 5).c_str(), 5) == 0 ) {
                isMe = true;
				sTmp = _T("* ") + sAuthor + aLine.substr(i+5);
			}
		}
	}
	sMyNick = client->getNick().c_str();
	if(BOOLSETTING(LOG_MAIN_CHAT)) {
		StringMap params;
		params["message"] = Text::fromT(sTmp);
		params["hub"] = client->getName();
		params["hubaddr"] = client->getAddressPort();
		params["mynick"] = client->getNick(); 
		LOG(LogManager::CHAT, params);
	}
	if(timeStamps) {
		ctrlClient.AppendText(Text::toT(sMyNick).c_str(), Text::toT("[" + Util::getShortTimeString() + "] ").c_str(), sTmp.c_str(), cf, sAuthor.c_str(), iAuthorLen, isMe, bUseEmo);
	} else {
		ctrlClient.AppendText(Text::toT(sMyNick).c_str(), _T(""), sTmp.c_str(), cf, sAuthor.c_str(), iAuthorLen, isMe, bUseEmo);
	}
	if (BOOLSETTING(TAB_DIRTY)) {
		setDirty();
	}
}

LRESULT HubFrame::onTabContextMenu(UINT /*uMsg*/, WPARAM /*wParam*/, LPARAM lParam, BOOL& /*bHandled*/) {
	POINT pt = { GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam) };        // location of mouse click 
	tabMenuShown = true;
	bool delseparator = false;
	prepareMenu(tabMenu, ::UserCommand::CONTEXT_HUB, Text::toT(client->getAddressPort()), client->getOp());
	if(!(tabMenu.GetMenuState(tabMenu.GetMenuItemCount()-1, MF_BYPOSITION) & MF_SEPARATOR)) {	
		tabMenu.AppendMenu(MF_SEPARATOR);
		delseparator = true;
	}
	tabMenu.AppendMenu(MF_STRING, IDC_CLOSE_WINDOW, CTSTRING(CLOSE));
	tabMenu.TrackPopupMenu(TPM_LEFTALIGN | TPM_BOTTOMALIGN | TPM_RIGHTBUTTON, pt.x, pt.y, m_hWnd);
	if(delseparator == true) {
		tabMenu.DeleteMenu(tabMenu.GetMenuItemCount()-1, MF_BYPOSITION);
	}
	tabMenu.DeleteMenu(tabMenu.GetMenuItemCount()-1, MF_BYPOSITION);
	cleanMenu(tabMenu);
	return TRUE;
}

LRESULT HubFrame::onContextMenu(UINT /*uMsg*/, WPARAM wParam, LPARAM lParam, BOOL& bHandled) {
	CRect rc;            // client area of window 
	POINT pt = { GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam) };        // location of mouse click
	tabMenuShown = false;
	CMenu Mnu;
	sSelectedUser = _T("");

	ctrlUsers.GetHeader().GetWindowRect(&rc);
		
	if(PtInRect(&rc, pt) && ShowUserList) {
		ctrlUsers.showMenu(pt);
		return TRUE;
	}

	if(reinterpret_cast<HWND>(wParam) == ctrlUsers && ShowUserList) { 
		if(ctrlUsers.GetSelectedCount() == 1) {
			if(pt.x == -1 && pt.y == -1) {
				WinUtil::getContextMenuPos(ctrlUsers, pt);
	
	
	
			}
			int i = -1;
			i = ctrlUsers.GetNextItem(i, LVNI_SELECTED);
			if(i >= 0) {
				sSelectedUser = Text::toT(((UserInfo*)ctrlUsers.getItemData(i))->user->getNick());
			}
		}

		if(PreparePopupMenu(&ctrlUsers, sSelectedUser, &Mnu)) {
			prepareMenu(Mnu, ::UserCommand::CONTEXT_CHAT, Text::toT(client->getAddressPort()), client->getOp());
			Mnu.TrackPopupMenu(TPM_LEFTALIGN | TPM_RIGHTBUTTON, pt.x, pt.y, m_hWnd);
			cleanMenu(Mnu);
			return TRUE;
		}
	}
	sSelectedUser = ChatCtrl::sTempSelectedUser;
	
	if(reinterpret_cast<HWND>(wParam) == ctrlClient) { 
		if(pt.x == -1 && pt.y == -1) {
			CRect erc;
			ctrlClient.GetRect(&erc);
			pt.x = erc.Width() / 2;
			pt.y = erc.Height() / 2;
			ctrlClient.ClientToScreen(&pt);
		}
		POINT ptCl = pt;
		ctrlClient.ScreenToClient(&ptCl); 
		bool boHitURL = ctrlClient.HitURL(ptCl);
		if(!boHitURL)
			sSelectedURL = _T("");

		if(PreparePopupMenu(&ctrlClient, sSelectedUser, &Mnu)) {
			if(sSelectedUser == _T("")) {
				Mnu.TrackPopupMenu(TPM_LEFTALIGN | TPM_RIGHTBUTTON, pt.x, pt.y, m_hWnd);
				return TRUE;
			} else {
				prepareMenu(Mnu, ::UserCommand::CONTEXT_CHAT, Text::toT(client->getAddressPort()), client->getOp());
				if(!(Mnu.GetMenuState(Mnu.GetMenuItemCount()-1, MF_BYPOSITION) & MF_SEPARATOR)) {	
					Mnu.AppendMenu(MF_SEPARATOR);
				}
				Mnu.AppendMenu(MF_STRING, ID_EDIT_CLEAR_ALL, CTSTRING(CLEAR));
				Mnu.TrackPopupMenu(TPM_LEFTALIGN | TPM_RIGHTBUTTON, pt.x, pt.y, m_hWnd);
				cleanMenu(Mnu);
				return TRUE;
			}
		}
	}
	return FALSE;
}

void HubFrame::runUserCommand(::UserCommand& uc) {
	if(!WinUtil::getUCParams(m_hWnd, uc, ucParams))
		return;

	ucParams["mynick"] = client->getNick();

	if(tabMenuShown) {
		client->escapeParams(ucParams);
		if(BOOLSETTING(USER_COMMANDS))
			addClientLine(_T("User command: ") + Text::toT(Util::formatParams(uc.getCommand(), ucParams)), WinUtil::m_ChatTextSystem);

		client->sendUserCmd(Util::formatParams(uc.getCommand(), ucParams));
	} else {
		int sel;
		UserInfo* u = NULL;
		if (sSelectedUser != _T("")) {
			sel = ctrlUsers.findItem(sSelectedUser);
			if ( sel >= 0 ) { 
				u = (UserInfo*)ctrlUsers.getItemData(sel);
				if(u->user->isOnline()) {
					StringMap tmp = ucParams;
					u->user->getParams(tmp);
					client->escapeParams(tmp); 
					if(BOOLSETTING(USER_COMMANDS))
						addClientLine(_T("User command: ") + Text::toT(Util::formatParams(uc.getCommand(), ucParams)), WinUtil::m_ChatTextSystem);

					client->sendUserCmd(Util::formatParams(uc.getCommand(), tmp));
				}
			}
		} else {
			sel = -1;
			while((sel = ctrlUsers.GetNextItem(sel, LVNI_SELECTED)) != -1) {
				u = (UserInfo*)ctrlUsers.getItemData(sel);
				if(u->user->isOnline()) {
					StringMap tmp = ucParams;
					u->user->getParams(tmp);
					client->escapeParams(tmp);
					if(BOOLSETTING(USER_COMMANDS))
						addClientLine(_T("User command: ") + Text::toT(Util::formatParams(uc.getCommand(), ucParams)), WinUtil::m_ChatTextSystem);

					client->sendUserCmd(Util::formatParams(uc.getCommand(), tmp));
				}
			}
		}
	}
	return;
};

void HubFrame::onTab() {
	if(BOOLSETTING(TAB_COMPLETION) && 
		(GetFocus() == ctrlMessage.m_hWnd) && 
		!(GetAsyncKeyState(VK_SHIFT) & 0x8000) ) 
	{
		int n = ctrlMessage.GetWindowTextLength();
		AutoArray<TCHAR> buf(n+1);
		ctrlMessage.GetWindowText(buf, n+1);
		tstring text(buf, n);
		string::size_type textStart = text.find_last_of(_T(" \n\t"));

		if(complete.empty()) {
			if(textStart != string::npos) {
				complete = text.substr(textStart + 1);
			} else {
				complete = text;
			}
			if(complete.empty()) {
				// Still empty, no text entered...
				return;
			}
			int y = ctrlUsers.GetItemCount();

			for(int x = 0; x < y; ++x)
				ctrlUsers.SetItemState(x, 0, LVNI_FOCUSED | LVNI_SELECTED);
		}

		if(textStart == string::npos)
			textStart = 0;
		else
			textStart++;

		int start = ctrlUsers.GetNextItem(-1, LVNI_FOCUSED) + 1;
		int i = start;
		int j = ctrlUsers.GetItemCount();

		bool firstPass = i < j;
		if(!firstPass)
			i = 0;
		while(firstPass || (!firstPass && i < start)) {
			UserInfo* ui = ctrlUsers.getItemData(i);
			const tstring& nick = ui->columns[UserInfo::COLUMN_NICK];
			bool found = (Util::strnicmp(nick, complete, complete.length()) == 0);
			tstring::size_type x = 0;
			if(!found) {
				// Check if there's one or more [ISP] tags to ignore...
				tstring::size_type y = 0;
				while(nick[y] == _T('[')) {
					x = nick.find(_T(']'), y);
					if(x != string::npos) {
						if(Util::strnicmp(nick.c_str() + x + 1, complete.c_str(), complete.length()) == 0) {
							found = true;
							break;
						}
					} else {
						break;
					}
					y = x + 1; // assuming that nick[y] == '\0' is legal
				}
			}
			if(found) {
				if((start - 1) != -1) {
					ctrlUsers.SetItemState(start - 1, 0, LVNI_SELECTED | LVNI_FOCUSED);
				}
				ctrlUsers.SetItemState(i, LVNI_FOCUSED | LVNI_SELECTED, LVNI_FOCUSED | LVNI_SELECTED);
				ctrlUsers.EnsureVisible(i, FALSE);
				ctrlMessage.SetSel(textStart, ctrlMessage.GetWindowTextLength(), TRUE);
				ctrlMessage.ReplaceSel(nick.c_str());
				return;
			}
			i++;
			if(i == j) {
				firstPass = false;
				i = 0;
			}
		}
	} else {
		HWND focus = GetFocus();

		if(focus == ctrlClient.m_hWnd) {
			ctrlMessage.SetFocus();
		} else if(focus == ctrlMessage.m_hWnd) {
			ctrlUsers.SetFocus();
		} else if(focus == ctrlUsers.m_hWnd) {
			ctrlFilter.SetFocus();
		} else if(focus == ctrlFilter.m_hWnd) {
			ctrlFilterSel.SetFocus();
		} else if(focus == ctrlFilterSel.m_hWnd) {
			ctrlClient.SetFocus();
		} 
	}
}

LRESULT HubFrame::onChar(UINT uMsg, WPARAM wParam, LPARAM /*lParam*/, BOOL& bHandled) {
	if (GetFocus() != ctrlMessage.m_hWnd) { // don't handle these keys unless the user is entering a message
		if (uMsg != WM_KEYDOWN) {
			if(wParam == VK_TAB) {
				bHandled = TRUE;
				return 0;
			}
		}
		if(wParam == VK_TAB) {
			onTab();
			return 0;
		}
		bHandled = FALSE;
		return 0;
	}
	if(!complete.empty() && wParam != VK_TAB && uMsg == WM_KEYDOWN)
		complete.clear();

	if (uMsg != WM_KEYDOWN) {
		switch(wParam) {
			case VK_RETURN:
				if( (GetKeyState(VK_CONTROL) & 0x8000) || (GetKeyState(VK_MENU) & 0x8000) ) {
					bHandled = FALSE;
				}
				break;
			case VK_TAB:
				bHandled = TRUE;
  				break;
  			default:
  				bHandled = FALSE;
				break;
			}
		return 0;
	}

	switch(wParam) {
		case VK_TAB:
			onTab();
			break;
		case VK_RETURN:
			if( (GetKeyState(VK_CONTROL) & 0x8000) || 
				(GetKeyState(VK_MENU) & 0x8000) ) {
					bHandled = FALSE;
				} else {
						onEnter();
					}
			break;
		case VK_UP:
			if ( (GetKeyState(VK_MENU) & 0x8000) ||	( ((GetKeyState(VK_CONTROL) & 0x8000) == 0) ^ (BOOLSETTING( SETTINGS_USE_CTRL_FOR_LINE_HISTORY ) == true) ) ) {
				//scroll up in chat command history
				//currently beyond the last command?
				if (curCommandPosition > 0) {
					//check whether current command needs to be saved
					if (curCommandPosition == prevCommands.size()) {
						auto_ptr<TCHAR> messageContents(new TCHAR[ctrlMessage.GetWindowTextLength()+2]);
						ctrlMessage.GetWindowText(messageContents.get(), ctrlMessage.GetWindowTextLength()+1);
						currentCommand = tstring(messageContents.get());
					}

					//replace current chat buffer with current command
					ctrlMessage.SetWindowText(prevCommands[--curCommandPosition].c_str());
				}
				// move cursor to end of line
				ctrlMessage.SetSel(ctrlMessage.GetWindowTextLength(), ctrlMessage.GetWindowTextLength());
			} else {
				bHandled = FALSE;
			}

			break;
		case VK_DOWN:
			if ( (GetKeyState(VK_MENU) & 0x8000) ||	( ((GetKeyState(VK_CONTROL) & 0x8000) == 0) ^ (BOOLSETTING( SETTINGS_USE_CTRL_FOR_LINE_HISTORY ) == true) ) ) {
				//scroll down in chat command history

				//currently beyond the last command?
				if (curCommandPosition + 1 < prevCommands.size()) {
					//replace current chat buffer with current command
					ctrlMessage.SetWindowText(prevCommands[++curCommandPosition].c_str());
				} else if (curCommandPosition + 1 == prevCommands.size()) {
					//revert to last saved, unfinished command

					ctrlMessage.SetWindowText(currentCommand.c_str());
					++curCommandPosition;
				}
				// move cursor to end of line
				ctrlMessage.SetSel(ctrlMessage.GetWindowTextLength(), ctrlMessage.GetWindowTextLength());
			} else {
				bHandled = FALSE;
			}

			break;
		case VK_PRIOR: // page up
			ctrlClient.SendMessage(WM_VSCROLL, SB_PAGEUP);

			break;
		case VK_NEXT: // page down
			ctrlClient.SendMessage(WM_VSCROLL, SB_PAGEDOWN);

			break;
		case VK_HOME:
			if (!prevCommands.empty() && (GetKeyState(VK_CONTROL) & 0x8000) || (GetKeyState(VK_MENU) & 0x8000)) {
				curCommandPosition = 0;
				
				auto_ptr<TCHAR> messageContents(new TCHAR[ctrlMessage.GetWindowTextLength()+2]);
				ctrlMessage.GetWindowText(messageContents.get(), ctrlMessage.GetWindowTextLength()+1);
				currentCommand = tstring(messageContents.get());

				ctrlMessage.SetWindowText(prevCommands[curCommandPosition].c_str());
			} else {
				bHandled = FALSE;
			}

			break;
		case VK_END:
			if ((GetKeyState(VK_CONTROL) & 0x8000) || (GetKeyState(VK_MENU) & 0x8000)) {
				curCommandPosition = prevCommands.size();

				ctrlMessage.SetWindowText(currentCommand.c_str());
			} else {
				bHandled = FALSE;
				}
				break;
		default:
			bHandled = FALSE;
	}
	return 0;
}

LRESULT HubFrame::onShowUsers(UINT /*uMsg*/, WPARAM wParam, LPARAM /*lParam*/, BOOL& bHandled) {
	bHandled = FALSE;
	if((wParam == BST_CHECKED)) {
		ShowUserList = true;
		User::NickMap& lst = client->lockUserList();
		ctrlUsers.SetRedraw(FALSE);
		for(User::NickIter i = lst.begin(); i != lst.end(); ++i) {
			updateUser(i->second);
		}
		client->unlockUserList();
		ctrlUsers.SetRedraw(TRUE);
		ctrlUsers.resort();
	} else {
		ShowUserList = false;
		clearUserList();
	}

	UpdateLayout(FALSE);
	return 0;
}

LRESULT HubFrame::onFollow(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
	if(!redirect.empty()) {
		string s, f;
		u_int16_t p = 411;
		Util::decodeUrl(Text::fromT(redirect), s, p, f);
		if(ClientManager::getInstance()->isConnected(s, p)) {
			addClientLine(TSTRING(REDIRECT_ALREADY_CONNECTED), WinUtil::m_ChatTextServer);
			return 0;
		}
		
		dcassert(frames.find(server) != frames.end());
		dcassert(frames[server] == this);
		frames.erase(server);
		server = redirect;
		frames[server] = this;

		// the client is dead, long live the client!
		client->removeListener(this);
		ClientManager::getInstance()->putClient(client);
		clearUserList();
		client = ClientManager::getInstance()->getClient(Text::fromT(server));

		RecentHubEntry r;
		r.setName("*");
		r.setDescription("***");
		r.setUsers("*");
		r.setShared("*");
		r.setServer(Text::fromT(redirect));
		FavoriteManager::getInstance()->addRecent(r);

		client->addListener(this);
		client->connect();
	}
	return 0;
}

LRESULT HubFrame::onEnterUsers(int /*idCtrl*/, LPNMHDR /* pnmh */, BOOL& /*bHandled*/) {
	int item = ctrlUsers.GetNextItem(-1, LVNI_FOCUSED);
	if(client->isConnected() && (item != -1)) {
		try {
			QueueManager::getInstance()->addList((ctrlUsers.getItemData(item))->user, QueueItem::FLAG_CLIENT_VIEW);
		} catch(const Exception& e) {
			addClientLine(Text::toT(e.getError()));
		}
	}
	return 0;
}

LRESULT HubFrame::onGetToolTip(int /*idCtrl*/, LPNMHDR pnmh, BOOL& /*bHandled*/) {
	NMTTDISPINFO* nm = (NMTTDISPINFO*)pnmh;
	lastLines.clear();
	for(TStringIter i = lastLinesList.begin(); i != lastLinesList.end(); ++i) {
		lastLines += *i;
		lastLines += _T("\r\n");
	}
	if(lastLines.size() > 2) {
		lastLines.erase(lastLines.size() - 2);
	}
	nm->lpszText = const_cast<TCHAR*>(lastLines.c_str());
	return 0;
}

void HubFrame::addClientLine(const tstring& aLine, bool inChat /* = true */) {
	tstring line = _T("[") + Text::toT(Util::getShortTimeString()) + _T("] ") + aLine;
    TCHAR* sLine = (TCHAR*)line.c_str();
    
    if(_tcslen(sLine) > 750) {
        sLine[750] = NULL;
    }

	ctrlStatus.SetText(0, sLine);
	while(lastLinesList.size() + 1 > MAX_CLIENT_LINES)
		lastLinesList.erase(lastLinesList.begin());
	lastLinesList.push_back(sLine);

	if(BOOLSETTING(TAB_DIRTY)) {
		setDirty();
	}
	
	if(BOOLSETTING(STATUS_IN_CHAT) && inChat) {
		addLine(_T("*** ") + aLine, WinUtil::m_ChatTextSystem);
	}
	if(BOOLSETTING(LOG_STATUS_MESSAGES)) {
		StringMap params;
		params["hub"] = client->getName();
		params["hubaddr"] = client->getAddressPort();
		params["message"] = Text::fromT(aLine);
		LOG(LogManager::STATUS, params);
	}
}

void HubFrame::closeDisconnected() {
	for(FrameIter i=frames.begin(); i!= frames.end(); ++i) {
		if (!(i->second->client->isConnected())) {
			i->second->PostMessage(WM_CLOSE);
		}
	}
};

void HubFrame::on(TimerManagerListener::Second, DWORD /*aTick*/) throw() {
	updateStatusBar();
	if(updateUsers) {
		updateUsers = false;
		PostMessage(WM_SPEAKER, UPDATE_USERS);
	}
}

void HubFrame::on(Connecting, Client*) throw() { 
	if(BOOLSETTING(SEARCH_PASSIVE) && (SETTING(CONNECTION_TYPE) == SettingsManager::CONNECTION_ACTIVE)) {
		addLine(TSTRING(ANTI_PASSIVE_SEARCH), WinUtil::m_ChatTextSystem);
	}
	speak(ADD_STATUS_LINE, STRING(CONNECTING_TO) + client->getAddressPort() + "...");
	speak(SET_WINDOW_TITLE, client->getAddressPort());
}
void HubFrame::on(Connected, Client*) throw() { 
	speak(CONNECTED);
}
void HubFrame::on(BadPassword, Client*) throw() { 
	client->setPassword(Util::emptyString);
}
void HubFrame::on(UserUpdated, Client*, const User::Ptr& user) throw() { 
	if(ShowUserList && !user->isSet(User::HIDDEN)) 
		speak(UPDATE_USER, user);
}
void HubFrame::on(UsersUpdated, Client*, const User::List& aList) throw() {
	Lock l(updateCS);
	updateList.reserve(aList.size());
	for(User::List::const_iterator i = aList.begin(); i != aList.end(); ++i) {
		if(!(*i)->isSet(User::HIDDEN))
			updateList.push_back(make_pair(*i, UPDATE_USERS));
	}
	if(!updateList.empty()) {
		PostMessage(WM_SPEAKER, UPDATE_USERS);
	}
}

void HubFrame::on(UserRemoved, Client*, const User::Ptr& user) throw() {
	if(ShowUserList) 
		speak(REMOVE_USER, user);
}

void HubFrame::on(Redirect, Client*, const string& line) throw() { 
	string s, f;
	u_int16_t p = 411;
	Util::decodeUrl(line, s, p, f);
	if(ClientManager::getInstance()->isConnected(s, p)) {
		speak(ADD_STATUS_LINE, STRING(REDIRECT_ALREADY_CONNECTED));
		return;
	}

	redirect = Text::toT(line);
	if(BOOLSETTING(AUTO_FOLLOW)) {
		PostMessage(WM_COMMAND, IDC_FOLLOW, 0);
	} else {
		speak(ADD_STATUS_LINE, STRING(PRESS_FOLLOW) + line);
	}
}
void HubFrame::on(Failed, Client*, const string& line) throw() { 
	speak(ADD_STATUS_LINE, line); 
	speak(DISCONNECTED); 
}
void HubFrame::on(GetPassword, Client*) throw() { 
	speak(GET_PASSWORD);
}
void HubFrame::on(HubUpdated, Client*) throw() { 
	speak(SET_WINDOW_TITLE, Util::validateMessage(client->getFullName(), true, false) + " (" + client->getAddressPort() + ")");
}
void HubFrame::on(Message, Client*, const char* line) throw() {
    char *temp, *chatline = (char *)line;
	if(SETTING(FILTER_MESSAGES)) {
		if((temp = strchr(chatline, '\n')) != NULL) {
            temp[0] = NULL;
            if(strstr(chatline, "was kicked by") != NULL) {
			    temp[0] = '\n';
                speak(ADD_SILENT_STATUS_LINE, Util::toDOS(line));
			    return;
            } else if((strstr(chatline, "is kicking") != NULL) && (strstr(chatline, "because:") != NULL)) {
                temp[0] = '\n';
                speak(ADD_SILENT_STATUS_LINE, Util::toDOS(line));
                return;
            }
            temp[0] = '\n';
        } else if(strstr(chatline, "was kicked by") != NULL) {
			speak(ADD_SILENT_STATUS_LINE, Util::toDOS(line));
			return;
        } else if((strstr(chatline, "is kicking") != NULL) && (strstr(chatline, "because:") != NULL)) {
			speak(ADD_SILENT_STATUS_LINE, Util::toDOS(line));
			return;
        }
	} else if((temp = strchr(chatline, '\n')) != NULL) {
        temp[0] = NULL;
        if((strstr(chatline, "is kicking") != NULL) && (strstr(chatline, "because:") != NULL) || 
		  (strstr(chatline, "was kicked by") != NULL)) {
              temp[0] = '\n';
		      speak(KICK_MSG, Util::toDOS(line));
		      return;
        }
        temp[0] = '\n';
	} else if((strstr(chatline, "is kicking") != NULL) && (strstr(chatline, "because:") != NULL) || 
		(strstr(chatline, "was kicked by") != NULL)) {
		speak(KICK_MSG, Util::toDOS(line));
        return;	
	}
	
	speak(ADD_CHAT_LINE, line);
}

void HubFrame::on(PrivateMessage, Client*, const User::Ptr& user, const string& line) throw() { 
	speak(PRIVATE_MESSAGE, user, line);
}
void HubFrame::on(NickTaken, Client*) throw() { 
	addLine(_T("*** ") + TSTRING(NICK_TAKEN), WinUtil::m_ChatTextServer);
	speak(DISCONNECTED);
}
void HubFrame::on(SearchFlood, Client*, const string& line) throw() {
	speak(ADD_STATUS_LINE, STRING(SEARCH_SPAM_FROM) + line);
}

void HubFrame::addClientLine(const tstring& aLine, CHARFORMAT2& cf, bool inChat /* = true */) {
	tstring line = _T("[") + Text::toT(Util::getShortTimeString()) + _T("] ") + aLine;

	ctrlStatus.SetText(0, line.c_str());
	while(lastLinesList.size() + 1 > MAX_CLIENT_LINES)
		lastLinesList.erase(lastLinesList.begin());
	lastLinesList.push_back(line);
	
	if (BOOLSETTING(TAB_DIRTY)) {
		setDirty();
	}
	
	if(BOOLSETTING(STATUS_IN_CHAT) && inChat) {
		addLine(_T("*** ") + aLine, cf);
	}
}

LRESULT HubFrame::onMouseMove(UINT /*uMsg*/, WPARAM /*wParam*/, LPARAM lParam, BOOL& bHandled) {
	RECT rc;
	POINT pt = { GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam) };        // location of mouse click
	// Get the bounding rectangle of the client area. 
	ctrlClient.GetClientRect(&rc);
	ctrlClient.ScreenToClient(&pt); 
	if (PtInRect(&rc, pt)) { 
		sSelectedURL = _T("");
  		return TRUE;
	}
	return 1;
}

LRESULT HubFrame::onRButtonUp(UINT /*uMsg*/, WPARAM /*wParam*/, LPARAM lParam, BOOL& bHandled) {
	RECT rc;
	POINT pt = { GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam) };        // location of mouse click

	// Get the bounding rectangle of the client area. 
	ctrlClient.GetClientRect(&rc);
	ctrlClient.ScreenToClient(&pt); 
	if (PtInRect(&rc, pt)) { 
		sSelectedURL = _T("");
	}
	return 1;
}

bool HubFrame::PreparePopupMenu(CWindow *pCtrl, tstring& sNick, CMenu *pMenu) {
	if (copyMenu.m_hMenu != NULL) {
		copyMenu.DestroyMenu();
		copyMenu.m_hMenu = NULL;
	}
	if (grantMenu.m_hMenu != NULL) {
		grantMenu.DestroyMenu();
		grantMenu.m_hMenu = NULL;
	}
	if (pMenu->m_hMenu != NULL) {
		pMenu->DestroyMenu();
		pMenu->m_hMenu = NULL;
	}

	copyMenu.CreatePopupMenu();
	copyMenu.AppendMenu(MF_STRING, IDC_COPY_NICK, CTSTRING(COPY_NICK));
	copyMenu.AppendMenu(MF_STRING, IDC_COPY_EXACT_SHARE, CTSTRING(COPY_EXACT_SHARE));
	copyMenu.AppendMenu(MF_STRING, IDC_COPY_DESCRIPTION, CTSTRING(COPY_DESCRIPTION));
	copyMenu.AppendMenu(MF_STRING, IDC_COPY_TAG, CTSTRING(COPY_TAG));
	copyMenu.AppendMenu(MF_STRING, IDC_COPY_EMAIL_ADDRESS, CTSTRING(COPY_EMAIL_ADDRESS));
	if (client->getOp()) {
		copyMenu.AppendMenu(MF_STRING, IDC_COPY_IP, CTSTRING(COPY_IP));
	}

	grantMenu.CreatePopupMenu();
	grantMenu.AppendMenu(MF_STRING, IDC_GRANTSLOT, CTSTRING(GRANT_EXTRA_SLOT));
	grantMenu.AppendMenu(MF_STRING, IDC_GRANTSLOT_HOUR, CTSTRING(GRANT_EXTRA_SLOT_HOUR));
	grantMenu.AppendMenu(MF_STRING, IDC_GRANTSLOT_DAY, CTSTRING(GRANT_EXTRA_SLOT_DAY));
	grantMenu.AppendMenu(MF_STRING, IDC_GRANTSLOT_WEEK, CTSTRING(GRANT_EXTRA_SLOT_WEEK));
	grantMenu.AppendMenu(MF_SEPARATOR);
	grantMenu.AppendMenu(MF_STRING, IDC_UNGRANTSLOT, CTSTRING(REMOVE_EXTRA_SLOT));

	pMenu->CreatePopupMenu();

	bool bIsChat = (pCtrl == ((CWindow*)&ctrlClient));
	if(sNick == _T("") && bIsChat == true) {
		if(ChatCtrl::sSelectedIP != _T("")) {
			pMenu->AppendMenu(MF_STRING, IDC_WHOIS_IP, (CTSTRING(WHO_IS) + ChatCtrl::sSelectedIP).c_str() );
			if(client->getOp()) {
				pMenu->AppendMenu(MF_SEPARATOR);
				pMenu->AppendMenu(MF_STRING, IDC_BAN_IP, (_T("!banip ") + ChatCtrl::sSelectedIP).c_str());
				pMenu->SetMenuDefaultItem(IDC_BAN_IP);
				pMenu->AppendMenu(MF_STRING, IDC_UNBAN_IP, (_T("!unban ") + ChatCtrl::sSelectedIP).c_str());
				pMenu->AppendMenu(MF_SEPARATOR);
			}
		}
		pMenu->AppendMenu(MF_STRING, ID_EDIT_COPY, CTSTRING(COPY));
		pMenu->AppendMenu(MF_STRING, IDC_COPY_ACTUAL_LINE,  CTSTRING(COPY_LINE));
		if(sSelectedURL != _T("")) 
  			pMenu->AppendMenu(MF_STRING, IDC_COPY_URL, CTSTRING(COPY_URL));
		pMenu->AppendMenu(MF_SEPARATOR);
		pMenu->AppendMenu(MF_STRING, ID_EDIT_SELECT_ALL, CTSTRING(SELECT_ALL));
		pMenu->AppendMenu(MF_STRING, ID_EDIT_CLEAR_ALL, CTSTRING(CLEAR));
		pMenu->AppendMenu(MF_SEPARATOR);
		pMenu->AppendMenu(MF_STRING, IDC_AUTOSCROLL_CHAT, CTSTRING(ASCROLL_CHAT));

		if(ctrlClient.GetAutoScroll())
			pMenu->CheckMenuItem(IDC_AUTOSCROLL_CHAT, MF_BYCOMMAND | MF_CHECKED);
	} else {
		if(sNick != _T("")) {
		    bool bIsMe = (sNick == Text::toT(sMyNick));
			// Jediny nick
			if(bIsChat) {
				tstring sTmp = TSTRING(USER) + _T(" ") + sNick;
				if(!BOOLSETTING(LOG_PRIVATE_CHAT)) {
					pMenu->AppendMenu(MF_STRING | MF_DISABLED, (UINT_PTR)0, sTmp.c_str());
				} else {
					pMenu->AppendMenu(MF_STRING, IDC_OPEN_USER_LOG, sTmp.c_str());
				}
				pMenu->AppendMenu(MF_SEPARATOR);
			}
			if(bIsChat == false && bIsMe == false && BOOLSETTING(LOG_PRIVATE_CHAT) == true) {
				pMenu->AppendMenu(MF_STRING, IDC_OPEN_USER_LOG, CTSTRING(OPEN_USER_LOG));
				pMenu->AppendMenu(MF_SEPARATOR);
			}
			if(bIsMe == false) {
				pMenu->AppendMenu(MF_STRING, IDC_PUBLIC_MESSAGE, CTSTRING(SEND_PUBLIC_MESSAGE));
				pMenu->AppendMenu(MF_STRING, IDC_PRIVATEMESSAGE, CTSTRING(SEND_PRIVATE_MESSAGE));
			}
			if(bIsChat) {
				pMenu->AppendMenu(MF_STRING, IDC_SELECT_USER, CTSTRING(SELECT_USER_LIST));
			}
			if(bIsMe == false) {
			   pMenu->AppendMenu(MF_SEPARATOR);
			}	   
			pMenu->AppendMenu(MF_POPUP, (UINT)(HMENU)copyMenu, CTSTRING(COPY));
			if(bIsMe == false) {
				pMenu->AppendMenu(MF_POPUP, (UINT)(HMENU)grantMenu, CTSTRING(GRANT_SLOTS_MENU));
				int i = ctrlUsers.findItem(sNick);
				if ( i >= 0 ) {
					UserInfo* ui = (UserInfo*)ctrlUsers.getItemData(i);
					if(client->getOp() || !ui->user->isSet(User::OP)) {
						pMenu->AppendMenu(MF_SEPARATOR);
						if(!ignoreList.count(sNick)) {
						      pMenu->AppendMenu(MF_STRING, IDC_IGNORE, CTSTRING(IGNORE_USER));
						} else {    
						      pMenu->AppendMenu(MF_STRING, IDC_UNIGNORE, CTSTRING(UNIGNORE_USER));
						}    
					}
				}
				pMenu->AppendMenu(MF_SEPARATOR);
				pMenu->AppendMenu(MF_STRING, IDC_GETLIST, CTSTRING(GET_FILE_LIST));
				pMenu->AppendMenu(MF_STRING, IDC_MATCH_QUEUE, CTSTRING(MATCH_QUEUE));
				pMenu->AppendMenu(MF_STRING, IDC_ADD_TO_FAVORITES, CTSTRING(ADD_TO_FAVORITES));
			}
		} else {
			// Muze byt vice nicku
			if(bIsChat == false) {
				// Pocet oznacenych
				int iCount = ctrlUsers.GetSelectedCount();
				TCHAR sTmp[256];
				_stprintf(sTmp, _T("%i user(s)"), iCount);
				pMenu->AppendMenu(MF_STRING | MF_DISABLED, (UINT_PTR)0, sTmp);
			}
			if (ctrlUsers.GetSelectedCount() <= 25) {
				pMenu->AppendMenu(MF_SEPARATOR);
				pMenu->AppendMenu(MF_STRING, IDC_PUBLIC_MESSAGE, CTSTRING(SEND_PUBLIC_MESSAGE));
			}
			pMenu->AppendMenu(MF_SEPARATOR);
			pMenu->AppendMenu(MF_STRING, IDC_GETLIST, CTSTRING(GET_FILE_LIST));
			pMenu->AppendMenu(MF_STRING, IDC_MATCH_QUEUE, CTSTRING(MATCH_QUEUE));
			pMenu->AppendMenu(MF_STRING, IDC_ADD_TO_FAVORITES, CTSTRING(ADD_TO_FAVORITES));
		}
		if(bIsChat) {
			switch(SETTING(CHAT_DBLCLICK)) {
		          case 0:
						pMenu->SetMenuDefaultItem(IDC_SELECT_USER);
						break;
		          case 1:
						pMenu->SetMenuDefaultItem(IDC_PUBLIC_MESSAGE);
						break;
		          case 2:
						pMenu->SetMenuDefaultItem(IDC_PRIVATEMESSAGE);
						break;
		          case 3:
						pMenu->SetMenuDefaultItem(IDC_GETLIST);
						break;
		          case 4:
						pMenu->SetMenuDefaultItem(IDC_MATCH_QUEUE);
						break;
		          case 6:
						pMenu->SetMenuDefaultItem(IDC_ADD_TO_FAVORITES);
						break;
			}    
		} else {    
			switch(SETTING(USERLIST_DBLCLICK)) {
		          case 0:
						pMenu->SetMenuDefaultItem(IDC_GETLIST);
						break;
		          case 1:
						pMenu->SetMenuDefaultItem(IDC_PUBLIC_MESSAGE);
						break;
		          case 2:
						pMenu->SetMenuDefaultItem(IDC_PRIVATEMESSAGE);
						break;
		          case 3:
						pMenu->SetMenuDefaultItem(IDC_MATCH_QUEUE);
						break;
		          case 5:
						pMenu->SetMenuDefaultItem(IDC_ADD_TO_FAVORITES);
						break;
			}
		}   		
	}
	return true;
}

LRESULT HubFrame::onCopyActualLine(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
	if((GetFocus() == ctrlClient.m_hWnd) && (ChatCtrl::sSelectedLine != _T(""))) {
		WinUtil::setClipboard(ChatCtrl::sSelectedLine);
	}
	return 0;
}

LRESULT HubFrame::onSelectUser(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
	if(sSelectedUser == _T("")) {
		// No nick selected
		return 0;
	}

	int pos = ctrlUsers.findItem(sSelectedUser);
	if ( pos == -1 ) {
		// User not found is list
		return 0;
	}

	int items = ctrlUsers.GetItemCount();
	ctrlUsers.SetRedraw(FALSE);
	for(int i = 0; i < items; ++i) {
		ctrlUsers.SetItemState(i, (i == pos) ? LVIS_SELECTED | LVIS_FOCUSED : 0, LVIS_SELECTED | LVIS_FOCUSED);
	}
	ctrlUsers.SetRedraw(TRUE);
	ctrlUsers.EnsureVisible(pos, FALSE);
	return 0;
}

LRESULT HubFrame::onPublicMessage(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
	int i = -1;
	CAtlString sUsers = "";
	CAtlString sText = "";

	if ( !client->isConnected() )
		return 0;

	if(sSelectedUser != _T("")) {
		sUsers = sSelectedUser.c_str();
	} else {
		while((i = ctrlUsers.GetNextItem(i, LVNI_SELECTED)) != -1) {
			if(sUsers.GetLength() > 0)
  				sUsers += ", ";
			sUsers += Text::toT(((UserInfo*)ctrlUsers.getItemData(i))->user->getNick()).c_str();
		}
	}

	int iSelBegin, iSelEnd;
	ctrlMessage.GetSel( iSelBegin, iSelEnd );
	ctrlMessage.GetWindowText( sText );

	if ( ( iSelBegin == 0 ) && ( iSelEnd == 0 ) ) {
		if ( sText.GetLength() == 0 ) {
			sUsers += ": ";
			ctrlMessage.SetWindowText( sUsers );
			ctrlMessage.SetFocus();
			ctrlMessage.SetSel( ctrlMessage.GetWindowTextLength(), ctrlMessage.GetWindowTextLength() );
		} else {
			sUsers += ": ";
			ctrlMessage.ReplaceSel( sUsers );
			ctrlMessage.SetFocus();
		}
	} else {
		sUsers += " ";
		ctrlMessage.ReplaceSel( sUsers );
		ctrlMessage.SetFocus();
	}
	return 0;
}

LRESULT HubFrame::onAutoScrollChat(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
	ctrlClient.SetAutoScroll( !ctrlClient.GetAutoScroll() );
	return 0;
}

LRESULT HubFrame::onBanIP(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
	if(ChatCtrl::sSelectedIP != _T("")) {
		tstring s = _T("!banip ") + ChatCtrl::sSelectedIP;
		if(BOOLSETTING(SERVER_COMMANDS))
			addClientLine(_T("Server command: ") + s, WinUtil::m_ChatTextSystem);

		client->hubMessage(Text::fromT(s));
	}
	return 0;
}

LRESULT HubFrame::onUnBanIP(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
	if(ChatCtrl::sSelectedIP != _T("")) {
		tstring s = _T("!unban ") + ChatCtrl::sSelectedIP;
		if(BOOLSETTING(SERVER_COMMANDS))
			addClientLine(_T("Server command: ") + s, WinUtil::m_ChatTextSystem);

		client->hubMessage(Text::fromT(s));
	}
	return 0;
}

LRESULT HubFrame::onCopyURL(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
	if(sSelectedURL != _T("")) {
		WinUtil::setClipboard(sSelectedURL);
	}
	return 0;
}

LRESULT HubFrame::onClientEnLink(int /*idCtrl*/, LPNMHDR pnmh, BOOL& /*bHandled*/) {
	ENLINK* pEL = (ENLINK*)pnmh;

	if ( pEL->msg == WM_LBUTTONUP ) {
		long lBegin = pEL->chrg.cpMin, lEnd = pEL->chrg.cpMax;
		TCHAR sURLTemp[INTERNET_MAX_URL_LENGTH];
		int iRet = ctrlClient.GetTextRange(lBegin, lEnd, sURLTemp);
		UNREFERENCED_PARAMETER(iRet);
		tstring sURL = sURLTemp;
		WinUtil::openLink(sURL);
	} else if(pEL->msg == WM_RBUTTONUP) {
		sSelectedURL = _T("");
		long lBegin = pEL->chrg.cpMin, lEnd = pEL->chrg.cpMax;
		TCHAR sURLTemp[INTERNET_MAX_URL_LENGTH];
		int iRet = ctrlClient.GetTextRange(lBegin, lEnd, sURLTemp);
		UNREFERENCED_PARAMETER(iRet);
		sSelectedURL = sURLTemp;

		ctrlClient.SetSel(lBegin, lEnd);
		ctrlClient.InvalidateRect(NULL);
		return 0;
	}
	return 0;
}

LRESULT HubFrame::onOpenUserLog(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {	
	StringMap params;
	UserInfo* ui = NULL;
	if(sSelectedUser != _T("")) {
		int k = ctrlUsers.findItem(sSelectedUser);
		if(k != -1) ui = ctrlUsers.getItemData(k);
	} else {
		int i = -1;
		if((i = ctrlUsers.GetNextItem(i, LVNI_SELECTED)) != -1) {
			ui = (UserInfo*)ctrlUsers.getItemData(i);
		}
	}
	if(ui == NULL) return 0;
	params["user"] = ui->user->getNick();
	params["hub"] = client->getName();
	params["mynick"] = ui->user->getClientNick();
	params["hubaddr"] = client->getAddressPort();
	tstring file = Text::toT(Util::validateFileName(SETTING(LOG_DIRECTORY) + Util::formatParams(SETTING(LOG_FILE_PRIVATE_CHAT), params)));
	if(File::existsFile(Text::fromT(file))) {
		ShellExecute(NULL, NULL, file.c_str(), NULL, NULL, SW_SHOWNORMAL);
	} else {
		MessageBox(CTSTRING(NO_LOG_FOR_USER),CTSTRING(NO_LOG_FOR_USER), MB_OK );	  
	}
	return 0;
}

LRESULT HubFrame::onOpenHubLog(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
	StringMap params;
	params["hub"] = client->getName();
	params["hubaddr"] = client->getAddressPort();
	params["mynick"] = client->getNick(); 
	tstring filename = Text::toT(Util::validateFileName(SETTING(LOG_DIRECTORY) + Util::formatParams(SETTING(LOG_FILE_MAIN_CHAT), params)));
	if(File::existsFile(Text::fromT(filename))){
		ShellExecute(NULL, NULL, filename.c_str(), NULL, NULL, SW_SHOWNORMAL);
	} else {
		MessageBox(CTSTRING(NO_LOG_FOR_HUB),CTSTRING(NO_LOG_FOR_HUB), MB_OK );	  
	}
	return 0;
}

LRESULT HubFrame::onWhoisIP(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
	if(ChatCtrl::sSelectedIP != _T("")) {
 		WinUtil::openLink(_T("http://www.ripe.net/perl/whois?form_type=simple&full_query_string=&searchtext=") + ChatCtrl::sSelectedIP);
 	}
	return 0;
}

LRESULT HubFrame::onStyleChange(UINT /*uMsg*/, WPARAM wParam, LPARAM /*lParam*/, BOOL& bHandled) {
	bHandled = FALSE;
	if((wParam & MK_LBUTTON) && ::GetCapture() == m_hWnd) {
		UpdateLayout(FALSE);
	}
	return 0;
}

LRESULT HubFrame::onStyleChanged(UINT /*uMsg*/, WPARAM /*wParam*/, LPARAM /*lParam*/, BOOL& bHandled) {
	bHandled = FALSE;
	UpdateLayout(FALSE);
	return 0;
}

LRESULT HubFrame::onFilterChar(UINT uMsg, WPARAM wParam, LPARAM /*lParam*/, BOOL& bHandled) {
	switch(wParam) {
		case VK_RETURN: {
			TCHAR *buf = new TCHAR[ctrlFilter.GetWindowTextLength()+1];
			ctrlFilter.GetWindowText(buf, ctrlFilter.GetWindowTextLength()+1);
			filterstring = Text::fromT(buf);
			free(buf);
			filtercolumn = ctrlFilterSel.GetCurSel();
			updateUserList();
			bHandled = FALSE;
			break; }
		case VK_TAB:
			onTab();
			break;
		default:
			bHandled = FALSE;
			break;
	}
	return 0;
}

void HubFrame::updateUserList() {
	Lock l(updateCS);

	ctrlUsers.SetRedraw(FALSE);
	ctrlUsers.DeleteAllItems();

	if(filterstring.empty()) {
		for(UserMap::iterator i = userMap.begin(); i != userMap.end(); ++i){
			if(i->second != NULL) {
				i->second->update();
				ctrlUsers.insertItem(i->second, WinUtil::getImage(i->second->user));
			}
		}
		ctrlUsers.SetRedraw(TRUE);
		return;
	}
	
	for(UserMap::iterator i = userMap.begin(); i != userMap.end(); ++i){
		if(i->second != NULL) {
			if(filterUser(i->first) == true) {
				ctrlUsers.insertItem(i->second, WinUtil::getImage(i->second->user));
			}
		}
	}
	ctrlUsers.SetRedraw(TRUE);
	ctrlUsers.resort();
} 

bool HubFrame::filterUser(const User::Ptr& aUser) {
	switch(filtercolumn) {
		case 0:
			//Nick
			if(Util::stristr(aUser->getNick().c_str(), filterstring.c_str()) != 0) {
				return true;
			}
			break;
		case 1:
			//Shared
			if(Util::stristr(Util::formatBytes(aUser->getBytesShared()).c_str(), filterstring.c_str()) != 0) {
				return true;
			}
			break;
		case 2:
			//Exact Shared
			if(Util::stristr(Util::toString(aUser->getBytesShared()).c_str(), filterstring.c_str()) != 0) {
				return true;
			}
			break;
		case 3:
			//Description
			if(Util::stristr(aUser->getDescription().c_str(), filterstring.c_str()) != 0) {
				return true;
			}
			break;
		case 4:
			//Tag
			if(Util::stristr(aUser->getTag().c_str(), filterstring.c_str()) != 0) {
				return true;
			}
			break;
		case 5:
			//Connection
			if(Util::stristr(aUser->getConnection().c_str(), filterstring.c_str()) != 0) {
				return true;
			}
			break;
		case 6:
			//Email
			if(Util::stristr(aUser->getEmail().c_str(), filterstring.c_str()) != 0) {
				return true;
			}
			break;
		case 7:
			//Client ID
			if(Util::stristr(aUser->getClientID().c_str(), filterstring.c_str()) != 0) {
				return true;
			}
			break;
		case 8:
			//Version
			if(Util::stristr(aUser->getVersion().c_str(), filterstring.c_str()) != 0) {
				return true;
			}
			break;
		case 9:
			//Mode
			if(Util::stristr(aUser->getMode().c_str(), filterstring.c_str()) != 0) {
				return true;
			}
			break;
		case 10:
			//Hubs
			if(Util::stristr(aUser->getHubs().c_str(), filterstring.c_str()) != 0) {
				return true;
			}
			break;
		case 11:
			//Slots
			if(Util::stristr(aUser->getSlots().c_str(), filterstring.c_str()) != 0) {
				return true;
			}
			break;
		case 12:
			//Limit
			if(Util::stristr(aUser->getUpload().c_str(), filterstring.c_str()) != 0) {
				return true;
			}
			break;
		case 13:
			//IP
			if(Util::stristr(aUser->getIp().c_str(), filterstring.c_str()) != 0) {
				return true;
			}
			break;
		default:
			return false;
			break;
	}
	return false;
}

void HubFrame::on(SettingsManagerListener::Save, SimpleXML* /*xml*/) throw() {
	if(ctrlUsers.GetBkColor() != WinUtil::bgColor) {
		ctrlClient.SetBackgroundColor(WinUtil::bgColor);
		ctrlUsers.SetBkColor(WinUtil::bgColor);
		ctrlUsers.SetTextBkColor(WinUtil::bgColor);
	}
	if(ctrlUsers.GetTextColor() != WinUtil::textColor) {
		ctrlUsers.SetTextColor(WinUtil::textColor);
	}
	RedrawWindow(NULL, NULL, RDW_ERASE | RDW_INVALIDATE | RDW_UPDATENOW | RDW_ALLCHILDREN);   
}

LRESULT HubFrame::onSizeMove(UINT /*uMsg*/, WPARAM /*wParam*/, LPARAM /*lParam*/, BOOL& bHandled) {
	bHandled = FALSE;
	ctrlClient.GoToEnd();
	return 0;
}

LRESULT HubFrame::onCopyHubName(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
    WinUtil::setClipboard(Text::toT(client->getName()));
	return 0;
}
    
LRESULT HubFrame::onCopyHubAddress(WORD /*wNotifyCode*/, WORD /*wID*/, HWND /*hWndCtl*/, BOOL& /*bHandled*/) {
    WinUtil::setClipboard(Text::toT(client->getAddressPort()));
	return 0;    
}    

LRESULT HubFrame::onCustomDraw(int /*idCtrl*/, LPNMHDR pnmh, BOOL& bHandled) {
	LPNMLVCUSTOMDRAW cd = (LPNMLVCUSTOMDRAW)pnmh;

	if(!BOOLSETTING(USERLIST_COLORING)) {
		bHandled = FALSE;
		return 0;
	}

	switch(cd->nmcd.dwDrawStage) {
		case CDDS_PREPAINT:
			return CDRF_NOTIFYITEMDRAW;

		case CDDS_ITEMPREPAINT: {
		    UserInfo* ui = (UserInfo*)cd->nmcd.lItemlParam; 
			if(ui->user->isFavoriteUser()){
				cd->clrText = SETTING(FAVORITE_COLOR);
			} else if(UploadManager::getInstance()->hasReservedSlot(ui->user)) {
				cd->clrText = SETTING(EXTRA_SLOT_COLOR);
			} else if(ignoreList.count(Text::toT(ui->user->getNick()))) {
				cd->clrText = SETTING(IGNORED_COLOR);
			} else if(ui->user->isSet(User::FIREBALL)) {
				cd->clrText = SETTING(FIREBALL_COLOR);
			} else if(ui->user->isSet(User::SERVER)) {
				cd->clrText = SETTING(SERVER_COLOR);
			} else if(ui->user->isSet(User::OP)) {
				cd->clrText = SETTING(OP_COLOR);
			} else if(ui->user->getMode() == "A" && !ui->user->isSet(User::PASSIVE)) {
				cd->clrText = SETTING(ACTIVE_COLOR);
			} else if(ui->user->isSet(User::PASSIVE) || ui->user->getMode() == "P" || ui->user->getMode() == "5") {
				cd->clrText = SETTING(PASIVE_COLOR);
			} else {
				cd->clrText = SETTING(NORMAL_COLOR);
			}
			return CDRF_NEWFONT | CDRF_NOTIFYSUBITEMDRAW;
		}

	default:
		return CDRF_DODEFAULT;
	}
}
	
/* Some parts Copyright (C) 2002-2005 Petr Kozelka, PPK at seznam dot cz */
